from __future__ import annotations
from typing import (TypeVar, Optional, Tuple, Callable, Any, Generic, List)
from .array import (fill, fold_back as fold_back_1, fold_back2 as fold_back2_1, iterate as iterate_1, scan_back as scan_back_1, try_find_back as try_find_back_1, try_find_index_back as try_find_index_back_1, permute as permute_1, map as map_1, chunk_by_size as chunk_by_size_1, pairwise as pairwise_1, windowed as windowed_1, split_into as split_into_1, transpose as transpose_1)
from .global_ import (SR_inputWasEmpty, SR_indexOutOfBounds, SR_keyNotFoundAlt, SR_differentLengths, IGenericAdder_1, IGenericAverager_1, SR_notEnoughElements, SR_inputMustBeNonNegative, SR_inputSequenceEmpty, SR_inputSequenceTooLong)
from .option import (some, value as value_1, default_arg)
from .reflection import (TypeInfo, option_type, record_type, class_type)
from .string import join
from .types import Record
from .util import (equals, structural_hash, compare, IEnumerator, to_iterator, get_enumerator, IDisposable, ignore, IEnumerable, is_array_like, uncurry, IEqualityComparer, IComparer)

_T = TypeVar("_T")

__29725_ = TypeVar("__29725_")

__29724_ = TypeVar("__29724_")

__29792_ = TypeVar("__29792_")

__29791_ = TypeVar("__29791_")

__29668_ = TypeVar("__29668_")

__29667_ = TypeVar("__29667_")

__29698_ = TypeVar("__29698_")

__29697_ = TypeVar("__29697_")

_A_ = TypeVar("_A_")

_STATE = TypeVar("_STATE")

_T1 = TypeVar("_T1")

_T2 = TypeVar("_T2")

_B_ = TypeVar("_B_")

_U = TypeVar("_U")

_T3 = TypeVar("_T3")

_RESULT = TypeVar("_RESULT")

_C_ = TypeVar("_C_")

def expr_71(gen0: TypeInfo) -> TypeInfo:
    return record_type("ListModule.FSharpList", [gen0], FSharpList, lambda: [("head", gen0), ("tail", option_type(FSharpList_reflection(gen0)))])


class FSharpList(Record, Generic[_T]):
    def __init__(self, head: _T, tail: Optional[FSharpList[_T]]) -> None:
        super().__init__()
        self.head = head
        self.tail = tail
    
    def __str__(self) -> str:
        xs : FSharpList[_T] = self
        return ("[" + join("; ", xs)) + "]"
    
    def __eq__(self, other: Any=None) -> bool:
        xs : FSharpList[_T] = self
        if xs is other:
            return True
        
        else: 
            def loop(xs_1_mut: FSharpList[__29725_], ys_1_mut: FSharpList[__29725_]) -> bool:
                while True:
                    (xs_1, ys_1) = (xs_1_mut, ys_1_mut)
                    match_value : Tuple[Optional[FSharpList[__29725_]], Optional[FSharpList[__29725_]]] = (xs_1.tail, ys_1.tail)
                    if match_value[0] is not None:
                        if match_value[1] is not None:
                            xt : FSharpList[__29725_] = match_value[0]
                            yt : FSharpList[__29725_] = match_value[1]
                            if equals(xs_1.head, ys_1.head):
                                xs_1_mut = xt
                                ys_1_mut = yt
                                continue
                            
                            else: 
                                return False
                            
                        
                        else: 
                            return False
                        
                    
                    elif match_value[1] is not None:
                        return False
                    
                    else: 
                        return True
                    
                    break
            
            loop : Callable[[FSharpList[__29724_], FSharpList[__29724_]], bool] = loop
            return loop(xs, other)
        
    
    def GetHashCode(self) -> int:
        xs : FSharpList[_T] = self
        def loop(i_mut: int, h_mut: int, xs_1_mut: FSharpList[_T]) -> int:
            while True:
                (i, h, xs_1) = (i_mut, h_mut, xs_1_mut)
                match_value : Optional[FSharpList[_T]] = xs_1.tail
                if match_value is not None:
                    t : FSharpList[_T] = match_value
                    if i > 18:
                        return h
                    
                    else: 
                        i_mut = i + 1
                        h_mut = ((h << 1) + structural_hash(xs_1.head)) + (631 * i)
                        xs_1_mut = t
                        continue
                    
                
                else: 
                    return h
                
                break
        
        return loop(0, 0, xs)
    
    def to_json(self, _key: str) -> Any:
        this : FSharpList[_T] = self
        return list(this)
    
    def CompareTo(self, other: Any=None) -> int:
        xs : FSharpList[_T] = self
        def loop(xs_1_mut: FSharpList[__29792_], ys_1_mut: FSharpList[__29792_]) -> int:
            while True:
                (xs_1, ys_1) = (xs_1_mut, ys_1_mut)
                match_value : Tuple[Optional[FSharpList[__29792_]], Optional[FSharpList[__29792_]]] = (xs_1.tail, ys_1.tail)
                if match_value[0] is not None:
                    if match_value[1] is not None:
                        xt : FSharpList[__29792_] = match_value[0]
                        yt : FSharpList[__29792_] = match_value[1]
                        c : int = compare(xs_1.head, ys_1.head) or 0
                        if c == 0:
                            xs_1_mut = xt
                            ys_1_mut = yt
                            continue
                        
                        else: 
                            return c
                        
                    
                    else: 
                        return 1
                    
                
                elif match_value[1] is not None:
                    return -1
                
                else: 
                    return 0
                
                break
        
        return loop(xs, other)
    
    def GetEnumerator(self) -> IEnumerator[_T]:
        xs : FSharpList[_T] = self
        return ListEnumerator_1__ctor_3002E699(xs)
    
    def __iter__(self) -> IEnumerator[_T]:
        return to_iterator(self.GetEnumerator())
    
    def System_Collections_IEnumerable_GetEnumerator(self) -> IEnumerator[Any]:
        xs : FSharpList[_T] = self
        return get_enumerator(xs)
    

FSharpList_reflection = expr_71

def expr_72(gen0: TypeInfo) -> TypeInfo:
    return class_type("ListModule.ListEnumerator`1", [gen0], ListEnumerator_1)


class ListEnumerator_1(IDisposable, Generic[_T]):
    def __init__(self, xs: FSharpList[_T]) -> None:
        self.xs = xs
        self.it = self.xs
        self.current = None
    
    def System_Collections_Generic_IEnumerator_00601_get_Current(self) -> _T:
        __ : ListEnumerator_1[_T] = self
        return __.current
    
    def System_Collections_IEnumerator_get_Current(self) -> Any:
        __ : ListEnumerator_1[_T] = self
        return __.current
    
    def System_Collections_IEnumerator_MoveNext(self) -> bool:
        __ : ListEnumerator_1[_T] = self
        match_value : Optional[FSharpList[_T]] = __.it.tail
        if match_value is not None:
            t : FSharpList[_T] = match_value
            __.current = __.it.head
            __.it = t
            return True
        
        else: 
            return False
        
    
    def System_Collections_IEnumerator_Reset(self) -> None:
        __ : ListEnumerator_1[_T] = self
        __.it = __.xs
        __.current = None
    
    def Dispose(self) -> None:
        pass
    

ListEnumerator_1_reflection = expr_72

def ListEnumerator_1__ctor_3002E699(xs: FSharpList[_T]) -> ListEnumerator_1[_T]:
    return ListEnumerator_1(xs)


def FSharpList_get_Empty() -> FSharpList[_T]:
    return FSharpList(None, None)


def FSharpList_Cons_305B8EAC(x: _T, xs: FSharpList[_T]) -> FSharpList[_T]:
    return FSharpList(x, xs)


def FSharpList__get_IsEmpty(xs: FSharpList[_T]) -> bool:
    return xs.tail is None


def FSharpList__get_Length(xs: FSharpList[_T]) -> int:
    def loop(i_mut: int, xs_1_mut: FSharpList[__29668_], xs: FSharpList[_T]=xs) -> int:
        while True:
            (i, xs_1) = (i_mut, xs_1_mut)
            match_value : Optional[FSharpList[__29668_]] = xs_1.tail
            if match_value is not None:
                i_mut = i + 1
                xs_1_mut = match_value
                continue
            
            else: 
                return i
            
            break
    
    return loop(0, xs)


def FSharpList__get_Head(xs: FSharpList[_T]) -> _T:
    match_value : Optional[FSharpList[_T]] = xs.tail
    if match_value is not None:
        return xs.head
    
    else: 
        raise Exception((SR_inputWasEmpty + "\\nParameter name: ") + "list")
    


def FSharpList__get_Tail(xs: FSharpList[_T]) -> FSharpList[_T]:
    match_value : Optional[FSharpList[_T]] = xs.tail
    if match_value is not None:
        return match_value
    
    else: 
        raise Exception((SR_inputWasEmpty + "\\nParameter name: ") + "list")
    


def FSharpList__get_Item_Z524259A4(xs: FSharpList[_T], index: int) -> _T:
    def loop(i_mut: int, xs_1_mut: FSharpList[__29698_], xs: FSharpList[_T]=xs, index: int=index) -> __29698_:
        while True:
            (i, xs_1) = (i_mut, xs_1_mut)
            match_value : Optional[FSharpList[__29698_]] = xs_1.tail
            if match_value is not None:
                if i == index:
                    return xs_1.head
                
                else: 
                    i_mut = i + 1
                    xs_1_mut = match_value
                    continue
                
            
            else: 
                raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
            
            break
    
    return loop(0, xs)


def empty() -> FSharpList[_A_]:
    return FSharpList_get_Empty()


def cons(x: _T, xs: FSharpList[_T]) -> FSharpList[_T]:
    return FSharpList_Cons_305B8EAC(x, xs)


def singleton(x: Optional[_A_]=None) -> FSharpList[_A_]:
    return FSharpList_Cons_305B8EAC(x, FSharpList_get_Empty())


def is_empty(xs: FSharpList[_T]) -> bool:
    return FSharpList__get_IsEmpty(xs)


def length(xs: FSharpList[_T]) -> int:
    return FSharpList__get_Length(xs)


def head(xs: FSharpList[_T]) -> _T:
    return FSharpList__get_Head(xs)


def try_head(xs: FSharpList[_T]) -> Optional[_T]:
    if FSharpList__get_IsEmpty(xs):
        return None
    
    else: 
        return some(FSharpList__get_Head(xs))
    


def tail(xs: FSharpList[_T]) -> FSharpList[_T]:
    return FSharpList__get_Tail(xs)


def try_last(xs_mut: FSharpList[_T]) -> Optional[_T]:
    while True:
        (xs,) = (xs_mut,)
        if FSharpList__get_IsEmpty(xs):
            return None
        
        else: 
            t : FSharpList[_T] = FSharpList__get_Tail(xs)
            if FSharpList__get_IsEmpty(t):
                return some(FSharpList__get_Head(xs))
            
            else: 
                xs_mut = t
                continue
            
        
        break


def last(xs: FSharpList[_T]) -> _T:
    match_value : Optional[_T] = try_last(xs)
    if match_value is None:
        raise Exception(SR_inputWasEmpty)
    
    else: 
        return value_1(match_value)
    


def compare_with(comparer: Callable[[_T, _T], int], xs: FSharpList[_T], ys: FSharpList[_T]) -> int:
    def loop(xs_1_mut: FSharpList[_T], ys_1_mut: FSharpList[_T], comparer: Callable[[_T, _T], int]=comparer, xs: FSharpList[_T]=xs, ys: FSharpList[_T]=ys) -> int:
        while True:
            (xs_1, ys_1) = (xs_1_mut, ys_1_mut)
            match_value : Tuple[bool, bool] = (FSharpList__get_IsEmpty(xs_1), FSharpList__get_IsEmpty(ys_1))
            if match_value[0]:
                if match_value[1]:
                    return 0
                
                else: 
                    return -1
                
            
            elif match_value[1]:
                return 1
            
            else: 
                c : int = comparer(FSharpList__get_Head(xs_1), FSharpList__get_Head(ys_1)) or 0
                if c == 0:
                    xs_1_mut = FSharpList__get_Tail(xs_1)
                    ys_1_mut = FSharpList__get_Tail(ys_1)
                    continue
                
                else: 
                    return c
                
            
            break
    
    return loop(xs, ys)


def to_array(xs: FSharpList[_T]) -> List[_T]:
    len_1 : int = FSharpList__get_Length(xs) or 0
    res : List[_T] = fill([0] * len_1, 0, len_1, None)
    def loop(i_mut: int, xs_1_mut: FSharpList[_T], xs: FSharpList[_T]=xs) -> None:
        while True:
            (i, xs_1) = (i_mut, xs_1_mut)
            if not FSharpList__get_IsEmpty(xs_1):
                res[i] = FSharpList__get_Head(xs_1)
                i_mut = i + 1
                xs_1_mut = FSharpList__get_Tail(xs_1)
                continue
            
            break
    
    loop(0, xs)
    return res


def fold(folder: Callable[[_STATE, _T], _STATE], state: _STATE, xs: FSharpList[_T]) -> _STATE:
    acc : _STATE = state
    xs_1 : FSharpList[_T] = xs
    while not FSharpList__get_IsEmpty(xs_1):
        acc = folder(acc, FSharpList__get_Head(xs_1))
        xs_1 = FSharpList__get_Tail(xs_1)
    return acc


def reverse(xs: FSharpList[_T]) -> FSharpList[_T]:
    def arrow_73(acc: FSharpList[_T], x: _T, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
        return FSharpList_Cons_305B8EAC(x, acc)
    
    return fold(arrow_73, FSharpList_get_Empty(), xs)


def fold_back(folder: Callable[[_T, _STATE], _STATE], xs: FSharpList[_T], state: _STATE) -> _STATE:
    return fold_back_1(folder, to_array(xs), state)


def fold_indexed(folder: Callable[[int, _STATE, _T], _STATE], state: _STATE, xs: FSharpList[_T]) -> _STATE:
    def loop(i_mut: int, acc_mut: _STATE, xs_1_mut: FSharpList[_T], folder: Callable[[int, _STATE, _T], _STATE]=folder, state: _STATE=state, xs: FSharpList[_T]=xs) -> _STATE:
        while True:
            (i, acc, xs_1) = (i_mut, acc_mut, xs_1_mut)
            if FSharpList__get_IsEmpty(xs_1):
                return acc
            
            else: 
                i_mut = i + 1
                acc_mut = folder(i, acc, FSharpList__get_Head(xs_1))
                xs_1_mut = FSharpList__get_Tail(xs_1)
                continue
            
            break
    
    return loop(0, state, xs)


def fold2(folder: Callable[[_STATE, _T1, _T2], _STATE], state: _STATE, xs: FSharpList[_T1], ys: FSharpList[_T2]) -> _STATE:
    acc : _STATE = state
    xs_1 : FSharpList[_T1] = xs
    ys_1 : FSharpList[_T2] = ys
    while (not FSharpList__get_IsEmpty(ys_1)) if (not FSharpList__get_IsEmpty(xs_1)) else False:
        acc = folder(acc, FSharpList__get_Head(xs_1), FSharpList__get_Head(ys_1))
        xs_1 = FSharpList__get_Tail(xs_1)
        ys_1 = FSharpList__get_Tail(ys_1)
    return acc


def fold_back2(folder: Callable[[_T1, _T2, _STATE], _STATE], xs: FSharpList[_T1], ys: FSharpList[_T2], state: _STATE) -> _STATE:
    return fold_back2_1(folder, to_array(xs), to_array(ys), state)


def unfold(gen: Callable[[_STATE], Optional[Tuple[_T, _STATE]]], state: _STATE) -> FSharpList[_T]:
    def loop(acc_mut: _STATE, node_mut: FSharpList[_T], gen: Callable[[_STATE], Optional[Tuple[_T, _STATE]]]=gen, state: _STATE=state) -> FSharpList[_T]:
        while True:
            (acc, node) = (acc_mut, node_mut)
            match_value : Optional[Tuple[_T, _STATE]] = gen(acc)
            if match_value is not None:
                acc_mut = match_value[1]
                def arrow_74(acc: _STATE=acc, node: FSharpList[_T]=node) -> FSharpList[_T]:
                    t : FSharpList[_T] = FSharpList(match_value[0], None)
                    node.tail = t
                    return t
                
                node_mut = arrow_74()
                continue
            
            else: 
                return node
            
            break
    
    root : FSharpList[_T] = FSharpList_get_Empty()
    node_1 : FSharpList[_T] = loop(state, root)
    t_2 : FSharpList[_T] = FSharpList_get_Empty()
    node_1.tail = t_2
    return FSharpList__get_Tail(root)


def iterate(action: Callable[[_A_], None], xs: FSharpList[_A_]) -> None:
    def arrow_75(unit_var0: None, x: _A_, action: Callable[[_A_], None]=action, xs: FSharpList[_A_]=xs) -> None:
        action(x)
    
    fold(arrow_75, None, xs)


def iterate2(action: Callable[[_A_, _B_], None], xs: FSharpList[_A_], ys: FSharpList[_B_]) -> None:
    def arrow_76(unit_var0: None, x: _A_, y: _B_, action: Callable[[_A_, _B_], None]=action, xs: FSharpList[_A_]=xs, ys: FSharpList[_B_]=ys) -> None:
        action(x, y)
    
    fold2(arrow_76, None, xs, ys)


def iterate_indexed(action: Callable[[int, _A_], None], xs: FSharpList[_A_]) -> None:
    def arrow_77(i: int, x: _A_, action: Callable[[int, _A_], None]=action, xs: FSharpList[_A_]=xs) -> int:
        action(i, x)
        return i + 1
    
    ignore(fold(arrow_77, 0, xs))


def iterate_indexed2(action: Callable[[int, _A_, _B_], None], xs: FSharpList[_A_], ys: FSharpList[_B_]) -> None:
    def arrow_78(i: int, x: _A_, y: _B_, action: Callable[[int, _A_, _B_], None]=action, xs: FSharpList[_A_]=xs, ys: FSharpList[_B_]=ys) -> int:
        action(i, x, y)
        return i + 1
    
    ignore(fold2(arrow_78, 0, xs, ys))


def to_seq(xs: FSharpList[_T]) -> IEnumerable[_T]:
    return xs


def of_array_with_tail(xs: List[_T], tail_1: FSharpList[_T]) -> FSharpList[_T]:
    res : FSharpList[_T] = tail_1
    for i in range(len(xs) - 1, 0 - 1, -1):
        res = FSharpList_Cons_305B8EAC(xs[i], res)
    return res


def of_array(xs: List[_T]) -> FSharpList[_T]:
    return of_array_with_tail(xs, FSharpList_get_Empty())


def of_seq(xs: IEnumerable[_T]) -> FSharpList[_T]:
    if is_array_like(xs):
        return of_array(xs)
    
    elif isinstance(xs, FSharpList):
        return xs
    
    else: 
        root : FSharpList[_T] = FSharpList_get_Empty()
        node : FSharpList[_T] = root
        with get_enumerator(xs) as enumerator:
            while enumerator.System_Collections_IEnumerator_MoveNext():
                x : _T = enumerator.System_Collections_Generic_IEnumerator_00601_get_Current()
                def arrow_79(xs: IEnumerable[_T]=xs) -> FSharpList[_T]:
                    xs_3 : FSharpList[_T] = node
                    t : FSharpList[_T] = FSharpList(x, None)
                    xs_3.tail = t
                    return t
                
                node = arrow_79()
        xs_5 : FSharpList[_T] = node
        t_2 : FSharpList[_T] = FSharpList_get_Empty()
        xs_5.tail = t_2
        return FSharpList__get_Tail(root)
    


def concat(lists: IEnumerable[FSharpList[_T]]) -> FSharpList[_T]:
    root : FSharpList[_T] = FSharpList_get_Empty()
    node : FSharpList[_T] = root
    def action(xs: FSharpList[_T], lists: IEnumerable[FSharpList[_T]]=lists) -> None:
        nonlocal node
        def arrow_80(acc: FSharpList[_T], x: _T, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
            t : FSharpList[_T] = FSharpList(x, None)
            acc.tail = t
            return t
        
        node = fold(arrow_80, node, xs)
    
    if is_array_like(lists):
        iterate_1(action, lists)
    
    elif isinstance(lists, FSharpList):
        iterate(action, lists)
    
    else: 
        with get_enumerator(lists) as enumerator:
            while enumerator.System_Collections_IEnumerator_MoveNext():
                action(enumerator.System_Collections_Generic_IEnumerator_00601_get_Current())
    
    xs_6 : FSharpList[_T] = node
    t_2 : FSharpList[_T] = FSharpList_get_Empty()
    xs_6.tail = t_2
    return FSharpList__get_Tail(root)


def scan(folder: Callable[[_STATE, _T], _STATE], state: _STATE, xs: FSharpList[_T]) -> FSharpList[_STATE]:
    root : FSharpList[_STATE] = FSharpList_get_Empty()
    def arrow_81(folder: Callable[[_STATE, _T], _STATE]=folder, state: _STATE=state, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
        t : FSharpList[_T] = FSharpList(state, None)
        root.tail = t
        return t
    
    node : FSharpList[_STATE] = arrow_81()
    acc : _STATE = state
    xs_3 : FSharpList[_T] = xs
    while not FSharpList__get_IsEmpty(xs_3):
        acc = folder(acc, FSharpList__get_Head(xs_3))
        def arrow_83(folder: Callable[[_STATE, _T], _STATE]=folder, state: _STATE=state, xs: FSharpList[_T]=xs) -> FSharpList[_STATE]:
            xs_4 : FSharpList[_STATE] = node
            def arrow_82(__unit: Any=None) -> FSharpList[_T]:
                t_2 : FSharpList[_T] = FSharpList(acc, None)
                xs_4.tail = t_2
                return t_2
            
            return arrow_82()
        
        node = arrow_83()
        xs_3 = FSharpList__get_Tail(xs_3)
    xs_6 : FSharpList[_STATE] = node
    t_4 : FSharpList[_STATE] = FSharpList_get_Empty()
    xs_6.tail = t_4
    return FSharpList__get_Tail(root)


def scan_back(folder: Callable[[_T, _STATE], _STATE], xs: FSharpList[_T], state: _STATE) -> FSharpList[_STATE]:
    return of_array(scan_back_1(folder, to_array(xs), state, None))


def append(xs: FSharpList[_T], ys: FSharpList[_T]) -> FSharpList[_T]:
    def arrow_84(acc: FSharpList[_T], x: _T, xs: FSharpList[_T]=xs, ys: FSharpList[_T]=ys) -> FSharpList[_T]:
        return FSharpList_Cons_305B8EAC(x, acc)
    
    return fold(arrow_84, ys, reverse(xs))


def collect(mapping: Callable[[_T], FSharpList[_U]], xs: FSharpList[_T]) -> FSharpList[_U]:
    root : FSharpList[_U] = FSharpList_get_Empty()
    node : FSharpList[_U] = root
    ys : FSharpList[_T] = xs
    while not FSharpList__get_IsEmpty(ys):
        zs : FSharpList[_U] = mapping(FSharpList__get_Head(ys))
        while not FSharpList__get_IsEmpty(zs):
            def arrow_86(mapping: Callable[[_T], FSharpList[_U]]=mapping, xs: FSharpList[_T]=xs) -> FSharpList[_U]:
                xs_1 : FSharpList[_U] = node
                def arrow_85(__unit: Any=None) -> FSharpList[_T]:
                    t : FSharpList[_T] = FSharpList(FSharpList__get_Head(zs), None)
                    xs_1.tail = t
                    return t
                
                return arrow_85()
            
            node = arrow_86()
            zs = FSharpList__get_Tail(zs)
        ys = FSharpList__get_Tail(ys)
    xs_3 : FSharpList[_U] = node
    t_2 : FSharpList[_U] = FSharpList_get_Empty()
    xs_3.tail = t_2
    return FSharpList__get_Tail(root)


def map_indexed(mapping: Callable[[int, _T], _U], xs: FSharpList[_T]) -> FSharpList[_U]:
    root : FSharpList[_U] = FSharpList_get_Empty()
    def folder(i: int, acc: FSharpList[_U], x: _T, mapping: Callable[[int, _T], _U]=mapping, xs: FSharpList[_T]=xs) -> FSharpList[_U]:
        def arrow_87(i: int=i, acc: FSharpList[_U]=acc, x: _T=x) -> FSharpList[_T]:
            t : FSharpList[_T] = FSharpList(mapping(i, x), None)
            acc.tail = t
            return t
        
        return arrow_87()
    
    node : FSharpList[_U] = fold_indexed(folder, root, xs)
    t_2 : FSharpList[_U] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def map(mapping: Callable[[_T], _U], xs: FSharpList[_T]) -> FSharpList[_U]:
    root : FSharpList[_U] = FSharpList_get_Empty()
    def folder(acc: FSharpList[_U], x: _T, mapping: Callable[[_T], _U]=mapping, xs: FSharpList[_T]=xs) -> FSharpList[_U]:
        def arrow_88(acc: FSharpList[_U]=acc, x: _T=x) -> FSharpList[_T]:
            t : FSharpList[_T] = FSharpList(mapping(x), None)
            acc.tail = t
            return t
        
        return arrow_88()
    
    node : FSharpList[_U] = fold(folder, root, xs)
    t_2 : FSharpList[_U] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def indexed(xs: FSharpList[_A_]) -> FSharpList[Tuple[int, _A_]]:
    def arrow_89(i: int, x: _A_, xs: FSharpList[_A_]=xs) -> Tuple[int, _A_]:
        return (i, x)
    
    return map_indexed(arrow_89, xs)


def map2(mapping: Callable[[_T1, _T2], _U], xs: FSharpList[_T1], ys: FSharpList[_T2]) -> FSharpList[_U]:
    root : FSharpList[_U] = FSharpList_get_Empty()
    def folder(acc: FSharpList[_U], x: _T1, y: _T2, mapping: Callable[[_T1, _T2], _U]=mapping, xs: FSharpList[_T1]=xs, ys: FSharpList[_T2]=ys) -> FSharpList[_U]:
        def arrow_90(acc: FSharpList[_U]=acc, x: _T1=x, y: _T2=y) -> FSharpList[_T]:
            t : FSharpList[_T] = FSharpList(mapping(x, y), None)
            acc.tail = t
            return t
        
        return arrow_90()
    
    node : FSharpList[_U] = fold2(folder, root, xs, ys)
    t_2 : FSharpList[_U] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def map_indexed2(mapping: Callable[[int, _T1, _T2], _U], xs: FSharpList[_T1], ys: FSharpList[_T2]) -> FSharpList[_U]:
    def loop(i_mut: int, acc_mut: FSharpList[_U], xs_1_mut: FSharpList[_T1], ys_1_mut: FSharpList[_T2], mapping: Callable[[int, _T1, _T2], _U]=mapping, xs: FSharpList[_T1]=xs, ys: FSharpList[_T2]=ys) -> FSharpList[_U]:
        while True:
            (i, acc, xs_1, ys_1) = (i_mut, acc_mut, xs_1_mut, ys_1_mut)
            if True if FSharpList__get_IsEmpty(xs_1) else FSharpList__get_IsEmpty(ys_1):
                return acc
            
            else: 
                i_mut = i + 1
                def arrow_91(i: int=i, acc: FSharpList[_U]=acc, xs_1: FSharpList[_T1]=xs_1, ys_1: FSharpList[_T2]=ys_1) -> FSharpList[_T]:
                    t : FSharpList[_T] = FSharpList(mapping(i, FSharpList__get_Head(xs_1), FSharpList__get_Head(ys_1)), None)
                    acc.tail = t
                    return t
                
                acc_mut = arrow_91()
                xs_1_mut = FSharpList__get_Tail(xs_1)
                ys_1_mut = FSharpList__get_Tail(ys_1)
                continue
            
            break
    
    root : FSharpList[_U] = FSharpList_get_Empty()
    node_1 : FSharpList[_U] = loop(0, root, xs, ys)
    t_2 : FSharpList[_U] = FSharpList_get_Empty()
    node_1.tail = t_2
    return FSharpList__get_Tail(root)


def map3(mapping: Callable[[_T1, _T2, _T3], _U], xs: FSharpList[_T1], ys: FSharpList[_T2], zs: FSharpList[_T3]) -> FSharpList[_U]:
    def loop(acc_mut: FSharpList[_U], xs_1_mut: FSharpList[_T1], ys_1_mut: FSharpList[_T2], zs_1_mut: FSharpList[_T3], mapping: Callable[[_T1, _T2, _T3], _U]=mapping, xs: FSharpList[_T1]=xs, ys: FSharpList[_T2]=ys, zs: FSharpList[_T3]=zs) -> FSharpList[_U]:
        while True:
            (acc, xs_1, ys_1, zs_1) = (acc_mut, xs_1_mut, ys_1_mut, zs_1_mut)
            if True if (True if FSharpList__get_IsEmpty(xs_1) else FSharpList__get_IsEmpty(ys_1)) else FSharpList__get_IsEmpty(zs_1):
                return acc
            
            else: 
                def arrow_92(acc: FSharpList[_U]=acc, xs_1: FSharpList[_T1]=xs_1, ys_1: FSharpList[_T2]=ys_1, zs_1: FSharpList[_T3]=zs_1) -> FSharpList[_T]:
                    t : FSharpList[_T] = FSharpList(mapping(FSharpList__get_Head(xs_1), FSharpList__get_Head(ys_1), FSharpList__get_Head(zs_1)), None)
                    acc.tail = t
                    return t
                
                acc_mut = arrow_92()
                xs_1_mut = FSharpList__get_Tail(xs_1)
                ys_1_mut = FSharpList__get_Tail(ys_1)
                zs_1_mut = FSharpList__get_Tail(zs_1)
                continue
            
            break
    
    root : FSharpList[_U] = FSharpList_get_Empty()
    node_1 : FSharpList[_U] = loop(root, xs, ys, zs)
    t_2 : FSharpList[_U] = FSharpList_get_Empty()
    node_1.tail = t_2
    return FSharpList__get_Tail(root)


def map_fold(mapping: Callable[[_STATE, _T], Tuple[_RESULT, _STATE]], state: _STATE, xs: FSharpList[_T]) -> Tuple[FSharpList[_RESULT], _STATE]:
    root : FSharpList[_RESULT] = FSharpList_get_Empty()
    def folder(tupled_arg: Tuple[FSharpList[_RESULT], _STATE], x: _T, mapping: Callable[[_STATE, _T], Tuple[_RESULT, _STATE]]=mapping, state: _STATE=state, xs: FSharpList[_T]=xs) -> Tuple[FSharpList[_RESULT], _STATE]:
        pattern_input : Tuple[_RESULT, _STATE] = mapping(tupled_arg[1], x)
        def arrow_93(tupled_arg: Tuple[FSharpList[_RESULT], _STATE]=tupled_arg, x: _T=x) -> FSharpList[_T]:
            t : FSharpList[_T] = FSharpList(pattern_input[0], None)
            tupled_arg[0].tail = t
            return t
        
        return (arrow_93(), pattern_input[1])
    
    pattern_input_1 : Tuple[FSharpList[_RESULT], _STATE] = fold(folder, (root, state), xs)
    t_2 : FSharpList[_RESULT] = FSharpList_get_Empty()
    pattern_input_1[0].tail = t_2
    return (FSharpList__get_Tail(root), pattern_input_1[1])


def map_fold_back(mapping: Callable[[_T, _STATE], Tuple[_RESULT, _STATE]], xs: FSharpList[_T], state: _STATE) -> Tuple[FSharpList[_RESULT], _STATE]:
    def arrow_94(acc: _STATE, x: _T, mapping: Callable[[_T, _STATE], Tuple[_RESULT, _STATE]]=mapping, xs: FSharpList[_T]=xs, state: _STATE=state) -> Tuple[_RESULT, _STATE]:
        return mapping(x, acc)
    
    return map_fold(arrow_94, state, reverse(xs))


def try_pick(f: Callable[[_T], Optional[_A_]], xs: FSharpList[_T]) -> Optional[_A_]:
    def loop(xs_1_mut: FSharpList[_T], f: Callable[[_T], Optional[_A_]]=f, xs: FSharpList[_T]=xs) -> Optional[_A_]:
        while True:
            (xs_1,) = (xs_1_mut,)
            if FSharpList__get_IsEmpty(xs_1):
                return None
            
            else: 
                match_value : Optional[_A_] = f(FSharpList__get_Head(xs_1))
                if match_value is None:
                    xs_1_mut = FSharpList__get_Tail(xs_1)
                    continue
                
                else: 
                    return match_value
                
            
            break
    
    return loop(xs)


def pick(f: Callable[[_A_], Optional[_B_]], xs: FSharpList[_A_]) -> _B_:
    match_value : Optional[_B_] = try_pick(f, xs)
    if match_value is None:
        raise Exception(SR_keyNotFoundAlt)
    
    else: 
        return value_1(match_value)
    


def try_find(f: Callable[[_A_], bool], xs: FSharpList[_A_]) -> Optional[_A_]:
    def arrow_95(x: Optional[_A_]=None, f: Callable[[_A_], bool]=f, xs: FSharpList[_A_]=xs) -> Optional[_A_]:
        return some(x) if f(x) else None
    
    return try_pick(arrow_95, xs)


def find(f: Callable[[_A_], bool], xs: FSharpList[_A_]) -> _A_:
    match_value : Optional[_A_] = try_find(f, xs)
    if match_value is None:
        raise Exception(SR_keyNotFoundAlt)
    
    else: 
        return value_1(match_value)
    


def try_find_back(f: Callable[[_A_], bool], xs: FSharpList[_A_]) -> Optional[_A_]:
    return try_find_back_1(f, to_array(xs))


def find_back(f: Callable[[_A_], bool], xs: FSharpList[_A_]) -> _A_:
    match_value : Optional[_A_] = try_find_back(f, xs)
    if match_value is None:
        raise Exception(SR_keyNotFoundAlt)
    
    else: 
        return value_1(match_value)
    


def try_find_index(f: Callable[[_T], bool], xs: FSharpList[_T]) -> Optional[int]:
    def loop(i_mut: int, xs_1_mut: FSharpList[_T], f: Callable[[_T], bool]=f, xs: FSharpList[_T]=xs) -> Optional[int]:
        while True:
            (i, xs_1) = (i_mut, xs_1_mut)
            if FSharpList__get_IsEmpty(xs_1):
                return None
            
            elif f(FSharpList__get_Head(xs_1)):
                return i
            
            else: 
                i_mut = i + 1
                xs_1_mut = FSharpList__get_Tail(xs_1)
                continue
            
            break
    
    return loop(0, xs)


def find_index(f: Callable[[_A_], bool], xs: FSharpList[_A_]) -> int:
    match_value : Optional[int] = try_find_index(f, xs)
    if match_value is None:
        raise Exception(SR_keyNotFoundAlt)
    
    else: 
        return match_value
    


def try_find_index_back(f: Callable[[_A_], bool], xs: FSharpList[_A_]) -> Optional[int]:
    return try_find_index_back_1(f, to_array(xs))


def find_index_back(f: Callable[[_A_], bool], xs: FSharpList[_A_]) -> int:
    match_value : Optional[int] = try_find_index_back(f, xs)
    if match_value is None:
        raise Exception(SR_keyNotFoundAlt)
    
    else: 
        return match_value
    


def try_item(n: int, xs: FSharpList[_T]) -> Optional[_T]:
    def loop(i_mut: int, xs_1_mut: FSharpList[_T], n: int=n, xs: FSharpList[_T]=xs) -> Optional[_T]:
        while True:
            (i, xs_1) = (i_mut, xs_1_mut)
            if FSharpList__get_IsEmpty(xs_1):
                return None
            
            elif i == n:
                return some(FSharpList__get_Head(xs_1))
            
            else: 
                i_mut = i + 1
                xs_1_mut = FSharpList__get_Tail(xs_1)
                continue
            
            break
    
    return loop(0, xs)


def item(n: int, xs: FSharpList[_T]) -> _T:
    return FSharpList__get_Item_Z524259A4(xs, n)


def filter(f: Callable[[_T], bool], xs: FSharpList[_T]) -> FSharpList[_T]:
    root : FSharpList[_T] = FSharpList_get_Empty()
    def folder(acc: FSharpList[_T], x: _T, f: Callable[[_T], bool]=f, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
        if f(x):
            t : FSharpList[_T] = FSharpList(x, None)
            acc.tail = t
            return t
        
        else: 
            return acc
        
    
    node : FSharpList[_T] = fold(folder, root, xs)
    t_2 : FSharpList[_T] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def partition(f: Callable[[_T], bool], xs: FSharpList[_T]) -> Tuple[FSharpList[_T], FSharpList[_T]]:
    pattern_input : Tuple[FSharpList[_T], FSharpList[_T]] = (FSharpList_get_Empty(), FSharpList_get_Empty())
    root2 : FSharpList[_T] = pattern_input[1]
    root1 : FSharpList[_T] = pattern_input[0]
    def folder(tupled_arg: Tuple[FSharpList[_T], FSharpList[_T]], f: Callable[[_T], bool]=f, xs: FSharpList[_T]=xs) -> Callable[[_T], Tuple[FSharpList[_T], FSharpList[_T]]]:
        lacc : FSharpList[_T] = tupled_arg[0]
        racc : FSharpList[_T] = tupled_arg[1]
        def arrow_98(x: Optional[_T]=None, tupled_arg: Tuple[FSharpList[_T], FSharpList[_T]]=tupled_arg) -> Tuple[FSharpList[_T], FSharpList[_T]]:
            def arrow_96(__unit: Any=None) -> FSharpList[_T]:
                t : FSharpList[_T] = FSharpList(x, None)
                lacc.tail = t
                return t
            
            def arrow_97(__unit: Any=None) -> FSharpList[_T]:
                t_2 : FSharpList[_T] = FSharpList(x, None)
                racc.tail = t_2
                return t_2
            
            return ((arrow_96(), racc)) if f(x) else ((lacc, arrow_97()))
        
        return arrow_98
    
    pattern_input_1 : Tuple[FSharpList[_T], FSharpList[_T]] = fold(uncurry(2, folder), (root1, root2), xs)
    t_4 : FSharpList[_T] = FSharpList_get_Empty()
    pattern_input_1[0].tail = t_4
    t_5 : FSharpList[_T] = FSharpList_get_Empty()
    pattern_input_1[1].tail = t_5
    return (FSharpList__get_Tail(root1), FSharpList__get_Tail(root2))


def choose(f: Callable[[_T], Optional[_T]], xs: FSharpList[_T]) -> FSharpList[_T]:
    root : FSharpList[_T] = FSharpList_get_Empty()
    def folder(acc: FSharpList[_T], x: _T, f: Callable[[_T], Optional[_T]]=f, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
        match_value : Optional[_T] = f(x)
        if match_value is None:
            return acc
        
        else: 
            t : FSharpList[_T] = FSharpList(value_1(match_value), None)
            acc.tail = t
            return t
        
    
    node : FSharpList[_T] = fold(folder, root, xs)
    t_2 : FSharpList[_T] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def contains(value: _T, xs: FSharpList[_T], eq: IEqualityComparer[Any]) -> bool:
    def arrow_99(v: Optional[_T]=None, value: _T=value, xs: FSharpList[_T]=xs, eq: IEqualityComparer[Any]=eq) -> bool:
        return eq.Equals(value, v)
    
    return try_find_index(arrow_99, xs) is not None


def initialize(n: int, f: Callable[[int], _T]) -> FSharpList[_T]:
    root : FSharpList[_T] = FSharpList_get_Empty()
    node : FSharpList[_T] = root
    for i in range(0, (n - 1) + 1, 1):
        def arrow_100(n: int=n, f: Callable[[int], _T]=f) -> FSharpList[_T]:
            xs : FSharpList[_T] = node
            t : FSharpList[_T] = FSharpList(f(i), None)
            xs.tail = t
            return t
        
        node = arrow_100()
    xs_2 : FSharpList[_T] = node
    t_2 : FSharpList[_T] = FSharpList_get_Empty()
    xs_2.tail = t_2
    return FSharpList__get_Tail(root)


def replicate(n: int, x: _A_) -> FSharpList[_A_]:
    def arrow_101(_arg1: int, n: int=n, x: _A_=x) -> _A_:
        return x
    
    return initialize(n, arrow_101)


def reduce(f: Callable[[_T, _T], _T], xs: FSharpList[_T]) -> _T:
    if FSharpList__get_IsEmpty(xs):
        raise Exception(SR_inputWasEmpty)
    
    else: 
        return fold(f, head(xs), tail(xs))
    


def reduce_back(f: Callable[[_T, _T], _T], xs: FSharpList[_T]) -> _T:
    if FSharpList__get_IsEmpty(xs):
        raise Exception(SR_inputWasEmpty)
    
    else: 
        return fold_back(f, tail(xs), head(xs))
    


def for_all(f: Callable[[_A_], bool], xs: FSharpList[_A_]) -> bool:
    def arrow_102(acc: bool, x: _A_, f: Callable[[_A_], bool]=f, xs: FSharpList[_A_]=xs) -> bool:
        return f(x) if acc else False
    
    return fold(arrow_102, True, xs)


def for_all2(f: Callable[[_A_, _B_], bool], xs: FSharpList[_A_], ys: FSharpList[_B_]) -> bool:
    def arrow_103(acc: bool, x: _A_, y: _B_, f: Callable[[_A_, _B_], bool]=f, xs: FSharpList[_A_]=xs, ys: FSharpList[_B_]=ys) -> bool:
        return f(x, y) if acc else False
    
    return fold2(arrow_103, True, xs, ys)


def exists(f: Callable[[_A_], bool], xs: FSharpList[_A_]) -> bool:
    return try_find_index(f, xs) is not None


def exists2(f_mut: Callable[[_T1, _T2], bool], xs_mut: FSharpList[_T1], ys_mut: FSharpList[_T2]) -> bool:
    while True:
        (f, xs, ys) = (f_mut, xs_mut, ys_mut)
        match_value : Tuple[bool, bool] = (FSharpList__get_IsEmpty(xs), FSharpList__get_IsEmpty(ys))
        (pattern_matching_result,) = (None,)
        if match_value[0]:
            if match_value[1]:
                pattern_matching_result = 0
            
            else: 
                pattern_matching_result = 2
            
        
        elif match_value[1]:
            pattern_matching_result = 2
        
        else: 
            pattern_matching_result = 1
        
        if pattern_matching_result == 0:
            return False
        
        elif pattern_matching_result == 1:
            if f(FSharpList__get_Head(xs), FSharpList__get_Head(ys)):
                return True
            
            else: 
                f_mut = f
                xs_mut = FSharpList__get_Tail(xs)
                ys_mut = FSharpList__get_Tail(ys)
                continue
            
        
        elif pattern_matching_result == 2:
            raise Exception((SR_differentLengths + "\\nParameter name: ") + "list2")
        
        break


def unzip(xs: FSharpList[Tuple[_A_, _B_]]) -> Tuple[FSharpList[_A_], FSharpList[_B_]]:
    def arrow_104(tupled_arg: Tuple[_A_, _B_], tupled_arg_1: Tuple[FSharpList[_A_], FSharpList[_B_]], xs: FSharpList[Tuple[_A_, _B_]]=xs) -> Tuple[FSharpList[_A_], FSharpList[_B_]]:
        return (FSharpList_Cons_305B8EAC(tupled_arg[0], tupled_arg_1[0]), FSharpList_Cons_305B8EAC(tupled_arg[1], tupled_arg_1[1]))
    
    return fold_back(arrow_104, xs, (FSharpList_get_Empty(), FSharpList_get_Empty()))


def unzip3(xs: FSharpList[Tuple[_A_, _B_, _C_]]) -> Tuple[FSharpList[_A_], FSharpList[_B_], FSharpList[_C_]]:
    def arrow_105(tupled_arg: Tuple[_A_, _B_, _C_], tupled_arg_1: Tuple[FSharpList[_A_], FSharpList[_B_], FSharpList[_C_]], xs: FSharpList[Tuple[_A_, _B_, _C_]]=xs) -> Tuple[FSharpList[_A_], FSharpList[_B_], FSharpList[_C_]]:
        return (FSharpList_Cons_305B8EAC(tupled_arg[0], tupled_arg_1[0]), FSharpList_Cons_305B8EAC(tupled_arg[1], tupled_arg_1[1]), FSharpList_Cons_305B8EAC(tupled_arg[2], tupled_arg_1[2]))
    
    return fold_back(arrow_105, xs, (FSharpList_get_Empty(), FSharpList_get_Empty(), FSharpList_get_Empty()))


def zip(xs: FSharpList[_A_], ys: FSharpList[_B_]) -> FSharpList[Tuple[_A_, _B_]]:
    def arrow_106(x: _A_, y: _B_, xs: FSharpList[_A_]=xs, ys: FSharpList[_B_]=ys) -> Tuple[_A_, _B_]:
        return (x, y)
    
    return map2(arrow_106, xs, ys)


def zip3(xs: FSharpList[_A_], ys: FSharpList[_B_], zs: FSharpList[_C_]) -> FSharpList[Tuple[_A_, _B_, _C_]]:
    def arrow_107(x: _A_, y: _B_, z: _C_, xs: FSharpList[_A_]=xs, ys: FSharpList[_B_]=ys, zs: FSharpList[_C_]=zs) -> Tuple[_A_, _B_, _C_]:
        return (x, y, z)
    
    return map3(arrow_107, xs, ys, zs)


def sort_with(comparer: Callable[[_T, _T], int], xs: FSharpList[_T]) -> FSharpList[_T]:
    arr : List[_T] = to_array(xs)
    arr.sort()
    return of_array(arr)


def sort(xs: FSharpList[_T], comparer: IComparer[_T]) -> FSharpList[_T]:
    def arrow_108(x: _T, y: _T, xs: FSharpList[_T]=xs, comparer: IComparer[_T]=comparer) -> int:
        return comparer.Compare(x, y)
    
    return sort_with(arrow_108, xs)


def sort_by(projection: Callable[[_T], _U], xs: FSharpList[_T], comparer: IComparer[_U]) -> FSharpList[_T]:
    def arrow_109(x: _T, y: _T, projection: Callable[[_T], _U]=projection, xs: FSharpList[_T]=xs, comparer: IComparer[_U]=comparer) -> int:
        return comparer.Compare(projection(x), projection(y))
    
    return sort_with(arrow_109, xs)


def sort_descending(xs: FSharpList[_T], comparer: IComparer[_T]) -> FSharpList[_T]:
    def arrow_110(x: _T, y: _T, xs: FSharpList[_T]=xs, comparer: IComparer[_T]=comparer) -> int:
        return comparer.Compare(x, y) * -1
    
    return sort_with(arrow_110, xs)


def sort_by_descending(projection: Callable[[_T], _U], xs: FSharpList[_T], comparer: IComparer[_U]) -> FSharpList[_T]:
    def arrow_111(x: _T, y: _T, projection: Callable[[_T], _U]=projection, xs: FSharpList[_T]=xs, comparer: IComparer[_U]=comparer) -> int:
        return comparer.Compare(projection(x), projection(y)) * -1
    
    return sort_with(arrow_111, xs)


def sum(xs: FSharpList[_T], adder: IGenericAdder_1[_T]) -> _T:
    def arrow_112(acc: _T, x: _T, xs: FSharpList[_T]=xs, adder: IGenericAdder_1[_T]=adder) -> _T:
        return adder.Add(acc, x)
    
    return fold(arrow_112, adder.GetZero(), xs)


def sum_by(f: Callable[[_T], _U], xs: FSharpList[_T], adder: IGenericAdder_1[_U]) -> _U:
    def arrow_113(acc: _U, x: _T, f: Callable[[_T], _U]=f, xs: FSharpList[_T]=xs, adder: IGenericAdder_1[_U]=adder) -> _U:
        return adder.Add(acc, f(x))
    
    return fold(arrow_113, adder.GetZero(), xs)


def max_by(projection: Callable[[_T], _U], xs: FSharpList[_T], comparer: IComparer[_U]) -> _T:
    def arrow_114(x: _T, y: _T, projection: Callable[[_T], _U]=projection, xs: FSharpList[_T]=xs, comparer: IComparer[_U]=comparer) -> _T:
        return y if (comparer.Compare(projection(y), projection(x)) > 0) else x
    
    return reduce(arrow_114, xs)


def max(xs: FSharpList[_T], comparer: IComparer[_T]) -> _T:
    def arrow_115(x: _T, y: _T, xs: FSharpList[_T]=xs, comparer: IComparer[_T]=comparer) -> _T:
        return y if (comparer.Compare(y, x) > 0) else x
    
    return reduce(arrow_115, xs)


def min_by(projection: Callable[[_T], _U], xs: FSharpList[_T], comparer: IComparer[_U]) -> _T:
    def arrow_116(x: _T, y: _T, projection: Callable[[_T], _U]=projection, xs: FSharpList[_T]=xs, comparer: IComparer[_U]=comparer) -> _T:
        return x if (comparer.Compare(projection(y), projection(x)) > 0) else y
    
    return reduce(arrow_116, xs)


def min(xs: FSharpList[_T], comparer: IComparer[_T]) -> _T:
    def arrow_117(x: _T, y: _T, xs: FSharpList[_T]=xs, comparer: IComparer[_T]=comparer) -> _T:
        return x if (comparer.Compare(y, x) > 0) else y
    
    return reduce(arrow_117, xs)


def average(xs: FSharpList[_T], averager: IGenericAverager_1[_T]) -> _T:
    count : int = 0
    def folder(acc: _T, x: _T, xs: FSharpList[_T]=xs, averager: IGenericAverager_1[_T]=averager) -> _T:
        nonlocal count
        count = (count + 1) or 0
        return averager.Add(acc, x)
    
    total : _T = fold(folder, averager.GetZero(), xs)
    return averager.DivideByInt(total, count)


def average_by(f: Callable[[_T], _U], xs: FSharpList[_T], averager: IGenericAverager_1[_U]) -> _U:
    count : int = 0
    def arrow_118(acc: _U, x: _T, f: Callable[[_T], _U]=f, xs: FSharpList[_T]=xs, averager: IGenericAverager_1[_U]=averager) -> _U:
        nonlocal count
        count = (count + 1) or 0
        return averager.Add(acc, f(x))
    
    total : _U = fold(arrow_118, averager.GetZero(), xs)
    return averager.DivideByInt(total, count)


def permute(f: Callable[[int], int], xs: FSharpList[_T]) -> FSharpList[_T]:
    return of_array(permute_1(f, to_array(xs)))


def chunk_by_size(chunk_size: int, xs: FSharpList[_T]) -> FSharpList[FSharpList[_T]]:
    def mapping(xs_1: List[_T], chunk_size: int=chunk_size, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
        return of_array(xs_1)
    
    return of_array(map_1(mapping, chunk_by_size_1(chunk_size, to_array(xs)), None))


def all_pairs(xs: FSharpList[_T1], ys: FSharpList[_T2]) -> FSharpList[Tuple[_T1, _T2]]:
    root : FSharpList[Tuple[_T1, _T2]] = FSharpList_get_Empty()
    node : FSharpList[Tuple[_T1, _T2]] = root
    def arrow_122(x: Optional[_T1]=None, xs: FSharpList[_T1]=xs, ys: FSharpList[_T2]=ys) -> None:
        def arrow_121(y: Optional[_T2]=None) -> None:
            nonlocal node
            def arrow_120(__unit: Any=None) -> FSharpList[Tuple[_T1, _T2]]:
                xs_1 : FSharpList[Tuple[_T1, _T2]] = node
                def arrow_119(__unit: Any=None) -> FSharpList[_T]:
                    t : FSharpList[_T] = FSharpList((x, y), None)
                    xs_1.tail = t
                    return t
                
                return arrow_119()
            
            node = arrow_120()
        
        iterate(arrow_121, ys)
    
    iterate(arrow_122, xs)
    xs_3 : FSharpList[Tuple[_T1, _T2]] = node
    t_2 : FSharpList[Tuple[_T1, _T2]] = FSharpList_get_Empty()
    xs_3.tail = t_2
    return FSharpList__get_Tail(root)


def skip(count_mut: int, xs_mut: FSharpList[_T]) -> FSharpList[_T]:
    while True:
        (count, xs) = (count_mut, xs_mut)
        if count <= 0:
            return xs
        
        elif FSharpList__get_IsEmpty(xs):
            raise Exception((SR_notEnoughElements + "\\nParameter name: ") + "list")
        
        else: 
            count_mut = count - 1
            xs_mut = FSharpList__get_Tail(xs)
            continue
        
        break


def skip_while(predicate_mut: Callable[[_T], bool], xs_mut: FSharpList[_T]) -> FSharpList[_T]:
    while True:
        (predicate, xs) = (predicate_mut, xs_mut)
        if FSharpList__get_IsEmpty(xs):
            return xs
        
        elif not predicate(FSharpList__get_Head(xs)):
            return xs
        
        else: 
            predicate_mut = predicate
            xs_mut = FSharpList__get_Tail(xs)
            continue
        
        break


def take(count: int, xs: FSharpList[_T]) -> FSharpList[_T]:
    if count < 0:
        raise Exception((SR_inputMustBeNonNegative + "\\nParameter name: ") + "count")
    
    def loop(i_mut: int, acc_mut: FSharpList[_T], xs_1_mut: FSharpList[_T], count: int=count, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
        while True:
            (i, acc, xs_1) = (i_mut, acc_mut, xs_1_mut)
            if i <= 0:
                return acc
            
            elif FSharpList__get_IsEmpty(xs_1):
                raise Exception((SR_notEnoughElements + "\\nParameter name: ") + "list")
            
            else: 
                i_mut = i - 1
                def arrow_123(i: int=i, acc: FSharpList[_T]=acc, xs_1: FSharpList[_T]=xs_1) -> FSharpList[_T]:
                    t : FSharpList[_T] = FSharpList(FSharpList__get_Head(xs_1), None)
                    acc.tail = t
                    return t
                
                acc_mut = arrow_123()
                xs_1_mut = FSharpList__get_Tail(xs_1)
                continue
            
            break
    
    root : FSharpList[_T] = FSharpList_get_Empty()
    node : FSharpList[_T] = loop(count, root, xs)
    t_2 : FSharpList[_T] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def take_while(predicate: Callable[[_T], bool], xs: FSharpList[_T]) -> FSharpList[_T]:
    def loop(acc_mut: FSharpList[_T], xs_1_mut: FSharpList[_T], predicate: Callable[[_T], bool]=predicate, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
        while True:
            (acc, xs_1) = (acc_mut, xs_1_mut)
            if FSharpList__get_IsEmpty(xs_1):
                return acc
            
            elif not predicate(FSharpList__get_Head(xs_1)):
                return acc
            
            else: 
                def arrow_124(acc: FSharpList[_T]=acc, xs_1: FSharpList[_T]=xs_1) -> FSharpList[_T]:
                    t : FSharpList[_T] = FSharpList(FSharpList__get_Head(xs_1), None)
                    acc.tail = t
                    return t
                
                acc_mut = arrow_124()
                xs_1_mut = FSharpList__get_Tail(xs_1)
                continue
            
            break
    
    root : FSharpList[_T] = FSharpList_get_Empty()
    node : FSharpList[_T] = loop(root, xs)
    t_2 : FSharpList[_T] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def truncate(count: int, xs: FSharpList[_T]) -> FSharpList[_T]:
    def loop(i_mut: int, acc_mut: FSharpList[_T], xs_1_mut: FSharpList[_T], count: int=count, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
        while True:
            (i, acc, xs_1) = (i_mut, acc_mut, xs_1_mut)
            if i <= 0:
                return acc
            
            elif FSharpList__get_IsEmpty(xs_1):
                return acc
            
            else: 
                i_mut = i - 1
                def arrow_125(i: int=i, acc: FSharpList[_T]=acc, xs_1: FSharpList[_T]=xs_1) -> FSharpList[_T]:
                    t : FSharpList[_T] = FSharpList(FSharpList__get_Head(xs_1), None)
                    acc.tail = t
                    return t
                
                acc_mut = arrow_125()
                xs_1_mut = FSharpList__get_Tail(xs_1)
                continue
            
            break
    
    root : FSharpList[_T] = FSharpList_get_Empty()
    node : FSharpList[_T] = loop(count, root, xs)
    t_2 : FSharpList[_T] = FSharpList_get_Empty()
    node.tail = t_2
    return FSharpList__get_Tail(root)


def get_slice(start_index: Optional[int], end_index: Optional[int], xs: FSharpList[_T]) -> FSharpList[_T]:
    len_1 : int = length(xs) or 0
    start_index_1 : int = default_arg(start_index, 0) or 0
    end_index_1 : int = default_arg(end_index, len_1 - 1) or 0
    if start_index_1 < 0:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "startIndex")
    
    elif end_index_1 >= len_1:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "endIndex")
    
    elif end_index_1 < start_index_1:
        return FSharpList_get_Empty()
    
    else: 
        return take((end_index_1 - start_index_1) + 1, skip(start_index_1, xs))
    


def split_at(index: int, xs: FSharpList[_T]) -> Tuple[FSharpList[_T], FSharpList[_T]]:
    if index < 0:
        raise Exception((SR_inputMustBeNonNegative + "\\nParameter name: ") + "index")
    
    if index > FSharpList__get_Length(xs):
        raise Exception((SR_notEnoughElements + "\\nParameter name: ") + "index")
    
    return (take(index, xs), skip(index, xs))


def exactly_one(xs: FSharpList[_T]) -> _T:
    if FSharpList__get_IsEmpty(xs):
        raise Exception((SR_inputSequenceEmpty + "\\nParameter name: ") + "list")
    
    elif FSharpList__get_IsEmpty(FSharpList__get_Tail(xs)):
        return FSharpList__get_Head(xs)
    
    else: 
        raise Exception((SR_inputSequenceTooLong + "\\nParameter name: ") + "list")
    


def try_exactly_one(xs: FSharpList[_T]) -> Optional[_T]:
    if FSharpList__get_IsEmpty(FSharpList__get_Tail(xs)) if (not FSharpList__get_IsEmpty(xs)) else False:
        return some(FSharpList__get_Head(xs))
    
    else: 
        return None
    


def where(predicate: Callable[[_T], bool], xs: FSharpList[_T]) -> FSharpList[_T]:
    return filter(predicate, xs)


def pairwise(xs: FSharpList[_T]) -> FSharpList[Tuple[_T, _T]]:
    return of_array(pairwise_1(to_array(xs)))


def windowed(window_size: int, xs: FSharpList[_T]) -> FSharpList[FSharpList[_T]]:
    def mapping(xs_1: List[_T], window_size: int=window_size, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
        return of_array(xs_1)
    
    return of_array(map_1(mapping, windowed_1(window_size, to_array(xs)), None))


def split_into(chunks: int, xs: FSharpList[_T]) -> FSharpList[FSharpList[_T]]:
    def mapping(xs_1: List[_T], chunks: int=chunks, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
        return of_array(xs_1)
    
    return of_array(map_1(mapping, split_into_1(chunks, to_array(xs)), None))


def transpose(lists: IEnumerable[FSharpList[_T]]) -> FSharpList[FSharpList[_T]]:
    def mapping_1(xs_1: List[_T], lists: IEnumerable[FSharpList[_T]]=lists) -> FSharpList[_T]:
        return of_array(xs_1)
    
    def mapping(xs: FSharpList[_T], lists: IEnumerable[FSharpList[_T]]=lists) -> List[_T]:
        return to_array(xs)
    
    return of_array(map_1(mapping_1, transpose_1(map_1(mapping, list(lists), None), None), None))


def insert_at(index: int, y: _T, xs: FSharpList[_T]) -> FSharpList[_T]:
    i : int = -1
    is_done : bool = False
    def folder(acc: FSharpList[_T], x: _T, index: int=index, y: _T=y, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
        nonlocal i, is_done
        i = (i + 1) or 0
        if i == index:
            is_done = True
            return FSharpList_Cons_305B8EAC(x, FSharpList_Cons_305B8EAC(y, acc))
        
        else: 
            return FSharpList_Cons_305B8EAC(x, acc)
        
    
    result : FSharpList[_T] = fold(folder, FSharpList_get_Empty(), xs)
    def arrow_137(index: int=index, y: _T=y, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
    
    return reverse(result if is_done else (FSharpList_Cons_305B8EAC(y, result) if ((i + 1) == index) else arrow_137()))


def insert_many_at(index: int, ys: IEnumerable[_T], xs: FSharpList[_T]) -> FSharpList[_T]:
    i : int = -1
    is_done : bool = False
    ys_1 : FSharpList[_T] = of_seq(ys)
    def folder(acc: FSharpList[_T], x: _T, index: int=index, ys: IEnumerable[_T]=ys, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
        nonlocal i, is_done
        i = (i + 1) or 0
        if i == index:
            is_done = True
            return FSharpList_Cons_305B8EAC(x, append(ys_1, acc))
        
        else: 
            return FSharpList_Cons_305B8EAC(x, acc)
        
    
    result : FSharpList[_T] = fold(folder, FSharpList_get_Empty(), xs)
    def arrow_139(index: int=index, ys: IEnumerable[_T]=ys, xs: FSharpList[_T]=xs) -> FSharpList[_T]:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
    
    return reverse(result if is_done else (append(ys_1, result) if ((i + 1) == index) else arrow_139()))


def remove_at(index: int, xs: FSharpList[_T]) -> FSharpList[_T]:
    i : int = -1
    is_done : bool = False
    def f(_arg1: Optional[_T]=None, index: int=index, xs: FSharpList[_T]=xs) -> bool:
        nonlocal i, is_done
        i = (i + 1) or 0
        if i == index:
            is_done = True
            return False
        
        else: 
            return True
        
    
    ys : FSharpList[_T] = filter(f, xs)
    if not is_done:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
    
    return ys


def remove_many_at(index: int, count: int, xs: FSharpList[_T]) -> FSharpList[_T]:
    i : int = -1
    status : int = -1
    def f(_arg1: Optional[_T]=None, index: int=index, count: int=count, xs: FSharpList[_T]=xs) -> bool:
        nonlocal i, status
        i = (i + 1) or 0
        if i == index:
            status = 0
            return False
        
        elif i > index:
            if i < (index + count):
                return False
            
            else: 
                status = 1
                return True
            
        
        else: 
            return True
        
    
    ys : FSharpList[_T] = filter(f, xs)
    status_1 : int = (1 if (((i + 1) == (index + count)) if (status == 0) else False) else status) or 0
    if status_1 < 1:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + ("index" if (status_1 < 0) else "count"))
    
    return ys


def update_at(index: int, y: _T, xs: FSharpList[_T]) -> FSharpList[_T]:
    is_done : bool = False
    def mapping(i: int, x: _T, index: int=index, y: _T=y, xs: FSharpList[_T]=xs) -> _T:
        nonlocal is_done
        if i == index:
            is_done = True
            return y
        
        else: 
            return x
        
    
    ys : FSharpList[_T] = map_indexed(mapping, xs)
    if not is_done:
        raise Exception((SR_indexOutOfBounds + "\\nParameter name: ") + "index")
    
    return ys





from __future__ import annotations
from typing import (Any, List)
from ..fable_library.reflection import (TypeInfo as TypeInfo_1, string_type, class_type, record_type, array_type, unit_type, lambda_type, tuple_type, union_type)
from ..fable_library.types import (Record, Union)

def expr_0() -> TypeInfo_1:
    return record_type("Fable.SimpleJson.Python.RecordField", [], RecordField, lambda: [("FieldName", string_type), ("FieldType", TypeInfo_reflection()), ("PropertyInfo", class_type("System.Reflection.PropertyInfo"))])


class RecordField(Record):
    def __init__(self, FieldName: str, FieldType: TypeInfo, PropertyInfo: Any) -> None:
        super().__init__()
        self.FieldName = FieldName
        self.FieldType = FieldType
        self.PropertyInfo = PropertyInfo
    

RecordField_reflection = expr_0

def expr_1() -> TypeInfo_1:
    return record_type("Fable.SimpleJson.Python.UnionCase", [], UnionCase, lambda: [("CaseName", string_type), ("CaseTypes", array_type(TypeInfo_reflection())), ("Info", class_type("Microsoft.FSharp.Reflection.UnionCaseInfo"))])


class UnionCase(Record):
    def __init__(self, CaseName: str, CaseTypes: List[TypeInfo], Info: Any) -> None:
        super().__init__()
        self.CaseName = CaseName
        self.CaseTypes = CaseTypes
        self.Info = Info
    

UnionCase_reflection = expr_1

def expr_2() -> TypeInfo_1:
    return union_type("Fable.SimpleJson.Python.TypeInfo", [], TypeInfo, lambda: [[], [], [], [], [], [], [], [], [], [], [], [], [], [], [], [], [], [], [], [], [], [], [("Item", lambda_type(unit_type, class_type("System.Type")))], [("Item", lambda_type(unit_type, TypeInfo_reflection()))], [("Item", lambda_type(unit_type, TypeInfo_reflection()))], [("Item", lambda_type(unit_type, TypeInfo_reflection()))], [("Item", lambda_type(unit_type, TypeInfo_reflection()))], [("Item", lambda_type(unit_type, TypeInfo_reflection()))], [("Item", lambda_type(unit_type, TypeInfo_reflection()))], [("Item", lambda_type(unit_type, TypeInfo_reflection()))], [("Item", lambda_type(unit_type, array_type(TypeInfo_reflection())))], [("Item", lambda_type(unit_type, tuple_type(TypeInfo_reflection(), TypeInfo_reflection())))], [("Item", lambda_type(unit_type, tuple_type(TypeInfo_reflection(), TypeInfo_reflection(), class_type("System.Type"))))], [("Item", lambda_type(unit_type, TypeInfo_reflection()))], [("Item", lambda_type(unit_type, TypeInfo_reflection()))], [("Item", lambda_type(unit_type, array_type(TypeInfo_reflection())))], [("Item", lambda_type(unit_type, tuple_type(TypeInfo_reflection(), class_type("System.Type"))))], [("Item", lambda_type(unit_type, tuple_type(array_type(RecordField_reflection()), class_type("System.Type"))))], [("Item", lambda_type(unit_type, tuple_type(array_type(UnionCase_reflection()), class_type("System.Type"))))]])


class TypeInfo(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag : int = tag or 0
        self.fields : List[Any] = list(fields)
    
    @staticmethod
    def cases() -> List[str]:
        return ["Unit", "Char", "String", "UInt16", "UInt32", "UInt64", "Int32", "Bool", "Float32", "Float", "Decimal", "Short", "Long", "Byte", "SByte", "DateTime", "DateTimeOffset", "BigInt", "TimeSpan", "Guid", "Object", "Uri", "Any", "Async", "Promise", "Option", "List", "Set", "Array", "Seq", "Tuple", "Map", "Dictionary", "ResizeArray", "HashSet", "Func", "Enum", "Record", "Union"]
    

TypeInfo_reflection = expr_2




from __future__ import annotations
from datetime import timedelta
import datetime
from dateutil import tz as tz_1
from typing import (Any, Optional, Tuple)
from ...fable_library.date import create
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.string import (to_console, printf)
from ...fs_hafas_python.context import Profile

def expr_290() -> TypeInfo:
    return class_type("FsHafas.Extensions.DateTimeOffsetEx.DateTimeOffsetEx", None, DateTimeOffsetEx)


class DateTimeOffsetEx:
    def __init__(self, dt: Any, tzoffset_arg: Optional[int]=None, tz_arg: Optional[str]=None) -> None:
        self.dt = dt
        self.tzoffset_arg = tzoffset_arg
        self.tz_arg = tz_arg
    

DateTimeOffsetEx_reflection = expr_290

def DateTimeOffsetEx__ctor_26581FE8(dt: Any, tzoffset_arg: Optional[int]=None, tz_arg: Optional[str]=None) -> DateTimeOffsetEx:
    return DateTimeOffsetEx(dt, tzoffset_arg, tz_arg)


def DateTimeOffsetEx__get_DateTime(__: DateTimeOffsetEx) -> Any:
    return DateTimeOffsetEx__getDateTime(__)


def DateTimeOffsetEx__AddDays_5E38073B(__: DateTimeOffsetEx, days: float) -> DateTimeOffsetEx:
    return DateTimeOffsetEx__ctor_26581FE8(__.dt+timedelta(days=int(days)), __.tzoffset_arg, __.tz_arg)


def DateTimeOffsetEx__getDateTime(this: DateTimeOffsetEx) -> Any:
    match_value : Tuple[Optional[int], Optional[str]] = (this.tzoffset_arg, this.tz_arg)
    if match_value[0] is not None:
        tzoffset_arg : int = match_value[0] or 0
        return datetime.datetime((this.dt.year), (this.dt.month), (this.dt.day), (this.dt.hour), (this.dt.minute), (this.dt.second), tzinfo=(tz_1.tzoffset(None, tzoffset_arg)))
    
    elif match_value[1] is not None:
        tz : str = match_value[1]
        return datetime.datetime((this.dt.year), (this.dt.month), (this.dt.day), (this.dt.hour), (this.dt.minute), (this.dt.second), tzinfo=(tz_1.gettz(tz)))
    
    else: 
        return datetime.datetime((this.dt.year), (this.dt.month), (this.dt.day), (this.dt.hour), (this.dt.minute), (this.dt.second), tzinfo=(tz_1.gettz("Europe/Berlin")))
    


def parse_date_time_with_offset(profile: Profile, year: int, month: int, day: int, hour: int, minute: int, seconds: int, tz_offset: Optional[int]=None) -> DateTimeOffsetEx:
    try: 
        dt : Any = create(year, month, day, hour, minute, seconds)
        return DateTimeOffsetEx__ctor_26581FE8(dt, None, None) if (tz_offset is None) else DateTimeOffsetEx__ctor_26581FE8(dt, tz_offset * 60, None)
    
    except Exception as ex:
        arg10 : str = str(ex)
        to_console(printf("error parseDateTimeWithOffset: %s"))(arg10)
        raise Exception(str(ex))
    


def ToIsoString(dto: DateTimeOffsetEx) -> str:
    dt : Any = DateTimeOffsetEx__get_DateTime(dto)
    return dt.isoformat()





from __future__ import annotations

import functools
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast

from .types import FSharpRef, Record
from .types import Union as FsUnion
from .util import combine_hash_codes, equal_arrays_with

Constructor = Callable[..., Any]

EnumCase = Tuple[str, int]
FieldInfo = Tuple[str, "TypeInfo"]
PropertyInfo = FieldInfo


@dataclass
class CaseInfo:
    declaringType: TypeInfo
    tag: int
    name: str
    fields: List[FieldInfo]


@dataclass
class TypeInfo:
    fullname: str
    generics: Optional[List[TypeInfo]] = None
    construct: Optional[Constructor] = None
    parent: Optional[TypeInfo] = None
    fields: Optional[Callable[[], List[FieldInfo]]] = None
    cases: Optional[Callable[[], List[CaseInfo]]] = None
    enum_cases: Optional[List[EnumCase]] = None

    def __str__(self) -> str:
        return full_name(self)

    def __eq__(self, other: Any) -> bool:
        return equals(self, other)

    def __hash__(self) -> int:
        hashes = list(map(hash, self.generics if self.generics else []))
        hashes.append(hash(self.fullname))
        return combine_hash_codes(hashes)


def class_type(
    fullname: str,
    generics: Optional[List[TypeInfo]] = None,
    construct: Optional[Constructor] = None,
    parent: Optional[TypeInfo] = None,
) -> TypeInfo:
    return TypeInfo(fullname, generics, construct, parent)


def union_type(
    fullname: str,
    generics: List[TypeInfo],
    construct: Type[FsUnion],
    cases: Callable[[], List[List[FieldInfo]]],
) -> TypeInfo:
    def fn() -> List[CaseInfo]:
        caseNames: List[str] = construct.cases()

        def mapper(i: int, fields: List[FieldInfo]) -> CaseInfo:
            return CaseInfo(t, i, caseNames[i], fields)

        return [mapper(i, x) for i, x in enumerate(cases())]

    t: TypeInfo = TypeInfo(fullname, generics, construct, None, None, fn, None)
    return t


def lambda_type(argType: TypeInfo, returnType: TypeInfo):
    return TypeInfo("Microsoft.FSharp.Core.FSharpFunc`2", [argType, returnType])


def delegate_type(*generics: TypeInfo) -> TypeInfo:
    return TypeInfo("System.Func`%d" % len(generics), list(generics))


def record_type(
    fullname: str,
    generics: List[TypeInfo],
    construct: Constructor,
    fields: Callable[[], List[FieldInfo]],
) -> TypeInfo:
    return TypeInfo(fullname, generics, construct, fields=fields)


def anon_record_type(*fields: FieldInfo) -> TypeInfo:
    return TypeInfo("", None, None, None, lambda: list(fields))


def option_type(generic: TypeInfo) -> TypeInfo:
    return TypeInfo("Microsoft.FSharp.Core.FSharpOption`1", [generic])


def list_type(generic: TypeInfo) -> TypeInfo:
    return TypeInfo("Microsoft.FSharp.Collections.FSharpList`1", [generic])


def array_type(generic: TypeInfo) -> TypeInfo:
    return TypeInfo(generic.fullname + "[]", [generic])


def enum_type(fullname: str, underlyingType: TypeInfo, enumCases: List[EnumCase]) -> TypeInfo:
    return TypeInfo(fullname, [underlyingType], None, None, None, None, enumCases)


def tuple_type(*generics: TypeInfo) -> TypeInfo:
    return TypeInfo(fullname=f"System.Tuple`{len(generics)}", generics=list(generics))


obj_type: TypeInfo = TypeInfo(fullname="System.Object")
unit_type: TypeInfo = TypeInfo("Microsoft.FSharp.Core.Unit")
char_type: TypeInfo = TypeInfo("System.Char")
string_type: TypeInfo = TypeInfo("System.String")
bool_type: TypeInfo = TypeInfo("System.Boolean")
int8_type: TypeInfo = TypeInfo("System.SByte")
uint8_type: TypeInfo = TypeInfo("System.Byte")
int16_type: TypeInfo = TypeInfo("System.Int16")
uint16_type: TypeInfo = TypeInfo("System.UInt16")
int32_type: TypeInfo = TypeInfo("System.Int32")
uint32_type: TypeInfo = TypeInfo("System.UInt32")
float32_type: TypeInfo = TypeInfo("System.Single")
float64_type: TypeInfo = TypeInfo("System.Double")
decimal_type: TypeInfo = TypeInfo("System.Decimal")


def equals(t1: TypeInfo, t2: TypeInfo) -> bool:
    if not t1.fullname:
        return not t2.fullname and equal_arrays_with(
            get_record_elements(t1),
            get_record_elements(t2),
            lambda kv1, kv2: kv1[0] == kv2[0] and equals(kv1[1], kv2[1]),
        )

    return t1.fullname == t2.fullname and equal_arrays_with(t1.generics, t2.generics, equals)


def is_generic_type(t: TypeInfo) -> bool:
    return t.generics is not None and len(t.generics) > 0


def get_generic_type_definition(t: TypeInfo):
    return t if t.generics is None else TypeInfo(t.fullname, list(map(lambda _: obj_type, t.generics)))


def get_generics(t: TypeInfo) -> List[TypeInfo]:
    return t.generics if t.generics else []


def make_generic_type(t: TypeInfo, generics: List[TypeInfo]) -> TypeInfo:
    return TypeInfo(t.fullname, generics, t.construct, t.parent, t.fields, t.cases)


def create_instance(t: TypeInfo, consArgs: List[Any]) -> Any:
    # TODO: Check if consArgs length is same as t.construct?
    # (Arg types can still be different)
    if callable(t.construct):
        return t.construct(*(consArgs or []))
    else:
        raise ValueError(f"Cannot access constructor of {t.fullname}")


def get_value(propertyInfo: PropertyInfo, v: Any) -> Any:
    return getattr(v, str(propertyInfo[0]))


def name(info: Union[FieldInfo, TypeInfo, CaseInfo]) -> str:
    if isinstance(info, Tuple):
        return cast(FieldInfo, info)[0]

    elif isinstance(info, CaseInfo):
        return info.name

    else:
        i = info.fullname.rfind(".")
        return info.fullname if i == -1 else info.fullname[i + 1 :]


def full_name(t: TypeInfo) -> str:
    gen = t.generics if t.generics and not is_array(t) else []
    if len(gen):
        gen_ = ",".join([full_name(x) for x in gen])
        return f"{t.fullname}[{gen_}]"

    else:
        return t.fullname


def namespace(t: TypeInfo) -> str:
    i = t.fullname.rfind(".")
    return "" if i == -1 else t.fullname[0:i]


def is_array(t: TypeInfo) -> bool:
    return t.fullname.endswith("[]")


def is_enum(t: TypeInfo) -> bool:
    return t.enum_cases is not None and len(t.enum_cases) > 0


def is_subclass_of(t1: TypeInfo, t2: TypeInfo) -> bool:
    return t1.parent is not None and ((t1.parent == t2) or is_subclass_of(t1.parent, t2))


def is_erased_to_number(t: TypeInfo) -> bool:
    return is_enum(t) or t.fullname in [
        int8_type.fullname,
        uint8_type.fullname,
        int16_type.fullname,
        uint16_type.fullname,
        int32_type.fullname,
        uint32_type.fullname,
        float32_type.fullname,
        float64_type.fullname,
    ]


def is_instance_of_type(t: TypeInfo, o: Any) -> bool:
    if t.fullname == obj_type.fullname:
        return True

    if isinstance(o, bool):
        return t.fullname == bool_type.fullname

    if isinstance(o, str):
        return t.fullname == string_type.fullname

    if isinstance(o, (int, float)):
        return is_erased_to_number(t)

    if callable(o):
        return is_function(t)

    return t.construct is not None and isinstance(o, t.construct)


def is_record(t: Any) -> bool:
    return (t.fields is not None) if isinstance(t, TypeInfo) else isinstance(t, Record)


def is_tuple(t: TypeInfo) -> bool:
    return t.fullname.startswith("System.Tuple") and not is_array(t)


def is_union(t: Any) -> bool:
    if isinstance(t, TypeInfo):
        return t.cases is not None

    return isinstance(t, FsUnion)


# In .NET this is false for delegates
def is_function(t: TypeInfo) -> bool:
    return t.fullname == "Microsoft.FSharp.Core.FSharpFunc`2"


def get_element_type(t: TypeInfo) -> Optional[TypeInfo]:
    return (t.generics[0] if t.generics else None) if is_array(t) else None


def get_enum_underlying_type(t: TypeInfo):
    return t.generics[0] if t.generics else None


def get_enum_values(t: TypeInfo) -> List[int]:
    if is_enum(t) and t.enum_cases is not None:
        return [int(kv[1]) for kv in t.enum_cases]
    else:
        raise ValueError(f"{t.fullname} is not an enum type")


def get_enum_names(t: TypeInfo) -> List[str]:
    if is_enum(t) and t.enum_cases is not None:
        return [str(kv[0]) for kv in t.enum_cases]
    else:
        raise ValueError(f"{t.fullname} is not an enum type")


def get_enum_case(t: TypeInfo, v: Union[int, str]) -> EnumCase:
    if t.enum_cases is None:
        raise ValueError(f"{t.fullname} is not an enum type")

    if isinstance(v, str):
        for kv in t.enum_cases:
            if kv[0] == v:
                return kv

        raise ValueError(f"{v}' was not found in ${t.fullname}")

    for kv in t.enum_cases:
        if kv[1] == v:
            return kv

    # .NET returns the number even if it doesn't match any of the cases
    return ("", v)


def get_tuple_elements(t: TypeInfo) -> List[TypeInfo]:
    if is_tuple(t) and t.generics is not None:
        return t.generics
    else:
        raise ValueError(f"{t.fullname} is not a tuple type")


def get_function_elements(t: TypeInfo) -> List[TypeInfo]:
    if is_function(t) and t.generics is not None:
        gen = t.generics
        return [gen[0], gen[1]]
    else:
        raise ValueError(f"{t.fullname} is not an F# function type")


def parse_enum(t: TypeInfo, string: str) -> int:
    value: Optional[int]
    try:
        value = int(string)
    except Exception:
        value = None

    return int(get_enum_case(t, value if value else string)[1])


def try_parse_enum(t: TypeInfo, string: str, def_value: FSharpRef[int]) -> bool:
    try:
        def_value.contents = parse_enum(t, string)
        return True
    except Exception:
        return False


def get_enum_name(t: TypeInfo, v: int) -> str:
    return str(get_enum_case(t, v)[0])


def is_enum_defined(t: TypeInfo, v: Union[str, int]) -> bool:
    try:
        kv = get_enum_case(t, v)
        return kv[0] is not None and kv[0] != ""
    except Exception:
        # Supress error
        pass

    return False


def get_record_elements(t: TypeInfo) -> List[FieldInfo]:
    if t.fields is not None:
        return t.fields()
    else:
        raise ValueError(f"{t.fullname} is not an F# record type")


def get_record_fields(v: Any) -> List[str]:
    if isinstance(v, Dict):
        return list(cast(Dict[str, Any], v).values())

    return [getattr(v, k) for k in v.__dict__.keys()]


def get_record_field(v: Any, field: FieldInfo) -> Any:
    if not isinstance(field[0], str):  # type: ignore
        raise ValueError("Field not a string.")

    if isinstance(v, dict):
        return v[field[0]]
    return getattr(v, field[0])


def get_tuple_fields(v: Any) -> List:
    return v


def get_tuple_field(v: Tuple[Any, ...], i: int) -> Any:
    return v[i]


def make_record(t: TypeInfo, values: List[Any]) -> Dict[str, Any]:
    fields = get_record_elements(t)
    if len(fields) != len(values):
        raise ValueError(f"Expected an array of length {len(fields)} but got {len(values)}")

    if t.construct is not None:
        return t.construct(*values)

    def reducer(obj: Dict[str, Any], ifield: Tuple[int, FieldInfo]):
        i, field = ifield
        obj[field[0]] = values[i]
        return obj

    initial: Dict[str, Any] = {}
    return functools.reduce(reducer, enumerate(fields), initial)


def make_tuple(values: List[Any], _t: TypeInfo) -> Tuple[Any, ...]:
    return tuple(values)


def make_union(uci: CaseInfo, values: List[Any]) -> Any:
    expectedLength = len(uci.fields or [])
    if len(values) != expectedLength:
        raise ValueError(f"Expected an array of length {expectedLength} but got {len(values)}")

    return uci.declaringType.construct(uci.tag, *values) if uci.declaringType.construct else {}


def get_union_cases(t: TypeInfo) -> List[CaseInfo]:
    if t.cases and callable(t.cases):
        return t.cases()
    else:
        raise ValueError(f"{t.fullname} is not an F# union type")


def get_union_fields(v: Any, t: TypeInfo) -> List[Any]:
    cases = get_union_cases(t)
    case_ = cases[v.tag]
    if not case_:
        raise ValueError(f"Cannot find case {v.name} in union type")

    return [case_, list(v.fields)]


def get_union_case_fields(uci: CaseInfo) -> List[FieldInfo]:
    return uci.fields if uci.fields else []


def assert_union(x: Any) -> None:
    if not isinstance(x, FsUnion):
        raise Exception("Value is not an F# union type")


def get_case_name(x: Any) -> str:
    assert_union(x)
    return x.cases()[x.tag]



from __future__ import annotations
from abc import abstractmethod
from typing import (TypeVar, List, Protocol, Generic, Any)

_T = TypeVar("_T")

class Cons_1(Protocol, Generic[_T]):
    @abstractmethod
    def Allocate(self, len: int) -> List[_T]:
        ...
    

def Helpers_allocateArrayFromCons(cons: Cons_1[_T], len_1: int) -> List[_T]:
    if cons is None:
        return (list)([None]*len_1)
    
    else: 
        return cons([0]*len_1)
    


def Helpers_fillImpl(array: List[_T], value: _T, start: int, count: int) -> List[_T]:
    for i in range(0, (count - 1) + 1, 1):
        array[i + start] = value
    return array


def Helpers_spliceImpl(array: List[_T], start: int, delete_count: int) -> List[_T]:
    for _ in range(1, delete_count + 1, 1):
        array.pop(start)
    return array


def Helpers_indexOfImpl(array: List[_T], item: _T, start: int) -> Any:
    try: 
        return array.index(item, start)
    
    except Exception as ex:
        return -1
    





from __future__ import annotations
from array import array as array_3
from typing import (Optional, Tuple, Any, List, MutableSequence, Callable)
from ..fable_library.array import (map, append, choose, fold)
from ..fable_library.date import (create, now)
from ..fable_library.int32 import parse as parse_1
from ..fable_library.map import (FSharpMap__ContainsKey, FSharpMap__get_Item, FSharpMap__Add, FSharpMap__Remove, to_array, empty)
from ..fable_library.option import (map as map_1, to_array as to_array_1)
from ..fable_library.seq import fold as fold_1
from ..fable_library.string import (substring, to_console, printf)
from ..fable_library.types import Int32Array
from ..fable_library.util import compare_primitives
from ..fs_hafas_python.lib.transformations import (Default_Location, Default_Journey, Default_Journeys)
from ..fs_hafas_python.parse.arrival_or_departure import DEP
from ..fs_hafas_python.parse.common import get_element_at
from .context import (CommonData, Options, Profile, Context, Profile__get_parseCommon, Profile__get_parseLocations, Profile__get_parseMovement, Profile__get_parseJourney, Profile__get_parseTrip, Profile__get_parseWarning, Profile__get_parseDeparture, Profile__get_parseArrival, Profile__get_parseDateTime)
from .types_hafas_client import (Movement, Duration, Journey, Journeys, Trip, Line, Warning, Alternative, ServerInfo)
from .types_raw_hafas_client import (RawResult, RawCommon, RawLoc, RawJny, RawPos, RawOutCon, RawLine, RawDir, RawHim)

default_common_data : CommonData = CommonData([], [], [], [], [], [])

default_options : Options = Options(True, True, True, False, True, True, True, False)

def create_context(profile: Profile, opt: Options, res: RawResult) -> Context:
    return Context(profile, opt, default_common_data, res)


def parse_common(profile: Profile, opt: Options, common: Optional[RawCommon]=None, res: Optional[RawResult]=None) -> Optional[Context]:
    match_value : Tuple[Optional[RawResult], Optional[RawCommon]] = (res, common)
    if match_value[0] is not None:
        if match_value[1] is None:
            res_2 : RawResult = match_value[0]
            return create_context(profile, opt, res_2)
        
        else: 
            common_1 : RawCommon = match_value[1]
            res_1 : RawResult = match_value[0]
            ctx : Context = create_context(profile, opt, res_1)
            return Context(ctx.profile, ctx.opt, Profile__get_parseCommon(ctx.profile)(ctx)(common_1), ctx.res)
        
    
    else: 
        return None
    


def parse_location(loc_l: Optional[RawLoc]=None, ctx: Optional[Context]=None) -> Any:
    match_value : Tuple[Optional[Context], Optional[RawLoc]] = (ctx, loc_l)
    (pattern_matching_result, ctx_1, loc_l_1) = (None, None, None)
    if match_value[0] is not None:
        if match_value[1] is not None:
            pattern_matching_result = 0
            ctx_1 = match_value[0]
            loc_l_1 = match_value[1]
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        locs : List[Any] = Profile__get_parseLocations(ctx_1.profile)(ctx_1)([loc_l_1])
        return locs[0]
    
    elif pattern_matching_result == 1:
        return Default_Location
    


def parse_locations(loc_l: Optional[List[RawLoc]]=None, ctx: Optional[Context]=None) -> List[Any]:
    match_value : Tuple[Optional[List[RawLoc]], Optional[Context]] = (loc_l, ctx)
    (pattern_matching_result, ctx_1, loc_l_1) = (None, None, None)
    if match_value[0] is not None:
        if match_value[1] is not None:
            pattern_matching_result = 0
            ctx_1 = match_value[1]
            loc_l_1 = match_value[0]
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        return Profile__get_parseLocations(ctx_1.profile)(ctx_1)(loc_l_1)
    
    elif pattern_matching_result == 1:
        return []
    


def parse_locations_from_result(profile: Profile, loc_l: Optional[List[RawLoc]], options: Options, res: RawResult) -> List[Any]:
    return parse_locations(loc_l, parse_common(profile, options, res.common, res))


def parse_movements(jny_l: Optional[List[RawJny]]=None, ctx: Optional[Context]=None) -> List[Movement]:
    match_value : Tuple[Optional[Context], Optional[List[RawJny]]] = (ctx, jny_l)
    (pattern_matching_result, ctx_1, jny_l_1) = (None, None, None)
    if match_value[0] is not None:
        if match_value[1] is not None:
            pattern_matching_result = 0
            ctx_1 = match_value[0]
            jny_l_1 = match_value[1]
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        return map(Profile__get_parseMovement(ctx_1.profile)(ctx_1), jny_l_1, None)
    
    elif pattern_matching_result == 1:
        return []
    


def parse_movements_from_result(profile: Profile, jny_l: Optional[List[RawJny]], options: Options, res: RawResult) -> List[Movement]:
    return parse_movements(jny_l, parse_common(profile, options, res.common, res))


def add_to_map(m: Any, p: RawPos) -> Any:
    if FSharpMap__ContainsKey(m, p.dur):
        l : MutableSequence[int] = FSharpMap__get_Item(m, p.dur)
        return FSharpMap__Add(FSharpMap__Remove(m, p.dur), p.dur, append(l, array_3("i", [p.loc_x]), Int32Array))
    
    else: 
        return FSharpMap__Add(m, p.dur, array_3("i", [p.loc_x]))
    


def get_locations(ctx: Context, loc_xs: MutableSequence[int]) -> List[Any]:
    def chooser(x: Optional[Any]=None, ctx: Context=ctx, loc_xs: MutableSequence[int]=loc_xs) -> Optional[Any]:
        return x
    
    def mapping(locx: int, ctx: Context=ctx, loc_xs: MutableSequence[int]=loc_xs) -> Optional[Any]:
        return get_element_at(locx, ctx.common.locations)
    
    return choose(chooser, map(mapping, loc_xs, None), None)


def parse_durations(pos_l: List[RawPos], ctx: Optional[Context]=None) -> List[Duration]:
    if ctx is not None:
        ctx_1 : Context = ctx
        def mapping(tupled_arg: Tuple[int, MutableSequence[int]], pos_l: List[RawPos]=pos_l, ctx: Optional[Context]=ctx) -> Duration:
            return Duration(tupled_arg[0], get_locations(ctx_1, tupled_arg[1]))
        
        def arrow_473(m: Any, p: RawPos, pos_l: List[RawPos]=pos_l, ctx: Optional[Context]=ctx) -> Any:
            return add_to_map(m, p)
        
        class ObjectExpr475:
            @property
            def Compare(self) -> Any:
                def arrow_474(x: int, y: int) -> int:
                    return compare_primitives(x, y)
                
                return arrow_474
            
        return map(mapping, to_array(fold(arrow_473, empty(ObjectExpr475()), pos_l)), None)
    
    else: 
        return []
    


def parse_durations_from_result(profile: Profile, pos_l: List[RawPos], options: Options, res: RawResult) -> List[Duration]:
    return parse_durations(pos_l, parse_common(profile, options, res.common, res))


def parse_journey(out_con_l: Optional[List[RawOutCon]]=None, ctx: Optional[Context]=None) -> Journey:
    match_value : Tuple[Optional[Context], Optional[List[RawOutCon]]] = (ctx, out_con_l)
    (pattern_matching_result, ctx_2, out_con_l_2) = (None, None, None)
    if match_value[0] is not None:
        if match_value[1] is not None:
            def arrow_476(out_con_l: Optional[List[RawOutCon]]=out_con_l, ctx: Optional[Context]=ctx) -> bool:
                ctx_1 : Context = match_value[0]
                return len(match_value[1]) > 0
            
            if arrow_476():
                pattern_matching_result = 0
                ctx_2 = match_value[0]
                out_con_l_2 = match_value[1]
            
            else: 
                pattern_matching_result = 1
            
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        return Profile__get_parseJourney(ctx_2.profile)(ctx_2)(out_con_l_2[0])
    
    elif pattern_matching_result == 1:
        return Default_Journey
    


def parse_journeys_array(out_con_l: Optional[List[RawOutCon]]=None, ctx: Optional[Context]=None) -> List[Journey]:
    match_value : Tuple[Optional[Context], Optional[List[RawOutCon]]] = (ctx, out_con_l)
    (pattern_matching_result, ctx_1, out_con_l_1) = (None, None, None)
    if match_value[0] is not None:
        if match_value[1] is not None:
            pattern_matching_result = 0
            ctx_1 = match_value[0]
            out_con_l_1 = match_value[1]
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        def mapping(o: RawOutCon, out_con_l: Optional[List[RawOutCon]]=out_con_l, ctx: Optional[Context]=ctx) -> Journey:
            return Profile__get_parseJourney(ctx_1.profile)(ctx_1)(o)
        
        return map(mapping, out_con_l_1, None)
    
    elif pattern_matching_result == 1:
        return []
    


def parse_journeys_array_from_result(profile: Profile, out_con_l: Optional[List[RawOutCon]], options: Options, res: RawResult) -> List[Journey]:
    return parse_journeys_array(out_con_l, parse_common(profile, options, res.common, res))


def parse_journeys(out_con_l: Optional[List[RawOutCon]]=None, ctx: Optional[Context]=None) -> Journeys:
    match_value : Tuple[Optional[Context], Optional[List[RawOutCon]]] = (ctx, out_con_l)
    (pattern_matching_result, ctx_1, out_con_l_1) = (None, None, None)
    if match_value[0] is not None:
        if match_value[1] is not None:
            pattern_matching_result = 0
            ctx_1 = match_value[0]
            out_con_l_1 = match_value[1]
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        def mapping(o: RawOutCon, out_con_l: Optional[List[RawOutCon]]=out_con_l, ctx: Optional[Context]=ctx) -> Journey:
            return Profile__get_parseJourney(ctx_1.profile)(ctx_1)(o)
        
        def mapping_1(p: str, out_con_l: Optional[List[RawOutCon]]=out_con_l, ctx: Optional[Context]=ctx) -> int:
            return parse_1(p, 511, False, 32)
        
        return Journeys(ctx_1.res.out_ctx_scr_b, ctx_1.res.out_ctx_scr_f, map(mapping, out_con_l_1, None), map_1(mapping_1, ctx_1.res.planrt_ts))
    
    elif pattern_matching_result == 1:
        return Default_Journeys
    


def parse_journeys_from_result(profile: Profile, out_con_l: Optional[List[RawOutCon]], options: Options, res: RawResult) -> Journeys:
    return parse_journeys(out_con_l, parse_common(profile, options, res.common, res))


def parse_trip(journey: Optional[RawJny]=None, ctx: Optional[Context]=None) -> Trip:
    match_value : Tuple[Optional[Context], Optional[RawJny]] = (ctx, journey)
    (pattern_matching_result, ctx_1, journey_1) = (None, None, None)
    if match_value[0] is not None:
        if match_value[1] is not None:
            pattern_matching_result = 0
            ctx_1 = match_value[0]
            journey_1 = match_value[1]
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        return Profile__get_parseTrip(ctx_1.profile)(ctx_1)(journey_1)
    
    elif pattern_matching_result == 1:
        raise Exception("parseTrip failed")
    


def parse_trip_from_result(profile: Profile, journey: Optional[RawJny], options: Options, res: RawResult) -> Trip:
    return parse_trip(journey, parse_common(profile, options, res.common, res))


def parse_trips(journeys: Optional[List[RawJny]]=None, ctx: Optional[Context]=None) -> List[Trip]:
    match_value : Tuple[Optional[Context], Optional[List[RawJny]]] = (ctx, journeys)
    (pattern_matching_result, ctx_1, journeys_1) = (None, None, None)
    if match_value[0] is not None:
        if match_value[1] is not None:
            pattern_matching_result = 0
            ctx_1 = match_value[0]
            journeys_1 = match_value[1]
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        def mapping(j: RawJny, journeys: Optional[List[RawJny]]=journeys, ctx: Optional[Context]=ctx) -> Trip:
            return Profile__get_parseTrip(ctx_1.profile)(ctx_1)(j)
        
        return map(mapping, journeys_1, None)
    
    elif pattern_matching_result == 1:
        raise Exception("parseTrip failed")
    


def parse_line(l: RawLine, ctx: Context) -> Optional[Line]:
    dirl : List[RawDir]
    match_value : Optional[RawCommon] = ctx.res.common
    if match_value is None:
        dirl = []
    
    else: 
        match_value_1 : Optional[List[RawDir]] = match_value.dir_l
        dirl = [] if (match_value_1 is None) else match_value_1
    
    def mapping_1(d_2: RawDir, l: RawLine=l, ctx: Context=ctx) -> str:
        return d_2.txt
    
    def chooser(x: Optional[RawDir]=None, l: RawLine=l, ctx: Context=ctx) -> Optional[RawDir]:
        return x
    
    def mapping(d_1: int, l: RawLine=l, ctx: Context=ctx) -> Optional[RawDir]:
        return get_element_at(d_1, dirl)
    
    def folder(_arg1: MutableSequence[int], d: MutableSequence[int], l: RawLine=l, ctx: Context=ctx) -> MutableSequence[int]:
        return d
    
    directions : List[str] = map(mapping_1, choose(chooser, map(mapping, fold_1(folder, array_3("i", []), to_array_1(l.dir_ref_l)), None), None), None)
    match_value_2 : Optional[Line] = get_element_at(l.prod_x, ctx.common.lines)
    if match_value_2 is None:
        return None
    
    else: 
        line : Line = match_value_2
        return Line(line.type, line.id, line.name, line.admin_code, line.fahrt_nr, line.additional_name, line.product, line.public, line.mode, line.routes, line.operator, line.express, line.metro, line.night, line.nr, line.symbol, directions, line.product_name)
    


def parse_lines(lines: Optional[List[RawLine]]=None, ctx: Optional[Context]=None) -> List[Line]:
    match_value : Tuple[Optional[Context], Optional[List[RawLine]]] = (ctx, lines)
    if match_value[0] is not None:
        if match_value[1] is None:
            ctx_2 : Context = match_value[0]
            return []
        
        else: 
            ctx_1 : Context = match_value[0]
            lines_1 : List[RawLine] = match_value[1]
            def chooser(x: Optional[Line]=None, lines: Optional[List[RawLine]]=lines, ctx: Optional[Context]=ctx) -> Optional[Line]:
                return x
            
            def mapping(l: RawLine, lines: Optional[List[RawLine]]=lines, ctx: Optional[Context]=ctx) -> Optional[Line]:
                return parse_line(l, ctx_1)
            
            return choose(chooser, map(mapping, lines_1, None), None)
        
    
    else: 
        raise Exception("parseLines failed")
    


def parse_lines_from_result(profile: Profile, lines: Optional[List[RawLine]], options: Options, res: RawResult) -> List[Line]:
    return parse_lines(lines, parse_common(profile, options, res.common, res))


def parse_warnings(msg_l: Optional[List[RawHim]]=None, ctx: Optional[Context]=None) -> List[Warning]:
    match_value : Tuple[Optional[Context], Optional[List[RawHim]]] = (ctx, msg_l)
    (pattern_matching_result, ctx_1, msg_l_1) = (None, None, None)
    if match_value[0] is not None:
        if match_value[1] is not None:
            pattern_matching_result = 0
            ctx_1 = match_value[0]
            msg_l_1 = match_value[1]
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        def mapping(j: RawHim, msg_l: Optional[List[RawHim]]=msg_l, ctx: Optional[Context]=ctx) -> Warning:
            return Profile__get_parseWarning(ctx_1.profile)(ctx_1)(j)
        
        return map(mapping, msg_l_1, None)
    
    elif pattern_matching_result == 1:
        raise Exception("parseWarnings failed")
    


def parse_warnings_from_result(profile: Profile, msg_l: Optional[List[RawHim]], options: Options, res: RawResult) -> List[Warning]:
    return parse_warnings(msg_l, parse_common(profile, options, res.common, res))


def ParseIsoString(datetime: str) -> Any:
    return create(parse_1(substring(datetime, 0, 4), 511, False, 32), parse_1(substring(datetime, 5, 2), 511, False, 32), parse_1(substring(datetime, 8, 2), 511, False, 32), parse_1(substring(datetime, 11, 2), 511, False, 32), parse_1(substring(datetime, 14, 2), 511, False, 32), 0)


def parse_departures_arrivals(type: str, jny_l: Optional[List[RawJny]]=None, ctx: Optional[Context]=None) -> List[Alternative]:
    match_value_1 : Tuple[Optional[Context], Optional[List[RawJny]]] = (ctx, jny_l)
    (pattern_matching_result, ctx_1, jny_l_1) = (None, None, None)
    if match_value_1[0] is not None:
        if match_value_1[1] is not None:
            pattern_matching_result = 0
            ctx_1 = match_value_1[0]
            jny_l_1 = match_value_1[1]
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        try: 
            parse : Callable[[Context, RawJny], Alternative] = Profile__get_parseDeparture(ctx_1.profile) if (type == DEP) else Profile__get_parseArrival(ctx_1.profile)
            def mapping(jny: RawJny) -> Alternative:
                return parse(ctx_1)(jny)
            
            arr : List[Alternative] = map(mapping, jny_l_1, None)
            def projection(d: Alternative) -> Any:
                try: 
                    match_value : Optional[str] = d.when
                    return now() if (match_value is None) else ParseIsoString(match_value)
                
                except Exception as ex:
                    arg10 : str = str(ex)
                    to_console(printf("%s"))(arg10)
                    return now()
                
            
            return sorted(arr, key=projection)
        
        except Exception as ex_1:
            arg10_1 : str = str(ex_1)
            to_console(printf("%s"))(arg10_1)
            return []
        
    
    elif pattern_matching_result == 1:
        return []
    


def parse_departures_arrivals_from_result(profile: Profile, type: str, jny_l: Optional[List[RawJny]], options: Options, res: RawResult) -> List[Alternative]:
    return parse_departures_arrivals(type, jny_l, parse_common(profile, options, res.common, res))


def parse_server_info(res: Optional[RawResult]=None, ctx: Optional[Context]=None) -> ServerInfo:
    match_value : Tuple[Optional[Context], Optional[RawResult]] = (ctx, res)
    (pattern_matching_result, ctx_1, res_1) = (None, None, None)
    if match_value[0] is not None:
        if match_value[1] is not None:
            pattern_matching_result = 0
            ctx_1 = match_value[0]
            res_1 = match_value[1]
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        def arrow_477(res: Optional[RawResult]=res, ctx: Optional[Context]=ctx) -> Optional[str]:
            match_value_1 : Optional[str] = res_1.s_d
            if match_value_1 is None:
                return None
            
            else: 
                s_d : str = match_value_1
                return Profile__get_parseDateTime(ctx_1.profile)(ctx_1)(s_d)(res_1.s_t)(None)
            
        
        def mapping(p: str, res: Optional[RawResult]=res, ctx: Optional[Context]=ctx) -> int:
            return parse_1(p, 511, False, 32)
        
        return ServerInfo(res_1.hci_version, res_1.fp_b, res_1.fp_e, arrow_477(), map_1(mapping, res_1.planrt_ts))
    
    elif pattern_matching_result == 1:
        raise Exception("ServerInfo failed")
    





from __future__ import annotations
from abc import abstractmethod
from typing import (Protocol, List)

class VersionInfo(Protocol):
    @property
    @abstractmethod
    def major(self) -> int:
        ...
    
    @property
    @abstractmethod
    def micro(self) -> int:
        ...
    
    @property
    @abstractmethod
    def minor(self) -> int:
        ...
    
    @property
    @abstractmethod
    def releaselevel(self) -> str:
        ...
    
    @property
    @abstractmethod
    def serial(self) -> int:
        ...
    

class IExports(Protocol):
    @property
    @abstractmethod
    def argv(self) -> List[str]:
        ...
    
    @property
    @abstractmethod
    def byteorder(self) -> str:
        ...
    
    @property
    @abstractmethod
    def hexversion(self) -> int:
        ...
    
    @property
    @abstractmethod
    def maxsize(self) -> int:
        ...
    
    @property
    @abstractmethod
    def maxunicode(self) -> int:
        ...
    
    @property
    @abstractmethod
    def path(self) -> List[str]:
        ...
    
    @property
    @abstractmethod
    def platform(self) -> str:
        ...
    
    @property
    @abstractmethod
    def prefix(self) -> str:
        ...
    
    @property
    @abstractmethod
    def version(self) -> str:
        ...
    
    @property
    @abstractmethod
    def version_info(self) -> VersionInfo:
        ...
    




from __future__ import annotations
from typing import Optional
from ...fable_library.option import default_arg
from ...fs_hafas_python.context import Context
from ...fs_hafas_python.extra_types import Icon
from ...fs_hafas_python.types_raw_hafas_client import RawIco

def parse_icon(ctx: Context, i: RawIco) -> Optional[Icon]:
    match_value : Optional[str] = i.res
    (pattern_matching_result, res_1) = (None, None)
    if match_value is not None:
        if match_value != "Empty":
            pattern_matching_result = 0
            res_1 = match_value
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        return Icon(res_1, default_arg(default_arg(i.text, i.txt), i.txt_s))
    
    elif pattern_matching_result == 1:
        return None
    





from setuptools import setup
import pathlib
from setuptools import setup

setup(
    name='fshafas',
    version='0.0.10',
    description='python client for HAFAS public transport APIs',
    long_description=open('package.readme.md').read(),
    long_description_content_type="text/markdown",
    url='https://github.com/bergmannjg/fshafas',
    license='MIT',
    author="Jürgen Bergmann",
    author_email="jbergmann@posteo.de",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.8',
    ],
    packages=[
        'fshafas',
        'fshafas/fable_modules/fs_hafas_python',
        'fshafas/fable_modules/fs_hafas_python/format',
        'fshafas/fable_modules/fs_hafas_python/parse',
        'fshafas/fable_modules/fs_hafas_python/lib',
        'fshafas/fable_modules/fs_hafas_python/extensions',
        'fshafas/fable_modules/fs_hafas_profiles_python/db',
        'fshafas/fable_modules/fable_simple_json_python',
        'fshafas/fable_modules/fable_python/stdlib',
        'fshafas/fable_modules/fable_library',
    ],
    install_requires=[
        'requests',
        'python-slugify',
        'polyline',
        'jsonpickle'
    ],
    python_requires='>=3.8',
)



from __future__ import annotations
from typing import (Optional, List, Tuple, TypeVar, Any)
from ..fable_library.async_builder import (singleton, Async)
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.string import (to_console, printf)
from ..fs_hafas_python.extensions.raw_request_ex import encode
from ..fs_hafas_python.extensions.raw_response import decode
from ..fs_hafas_python.lib.request import (HttpClient__ctor, HttpClient__Dispose, HttpClient__PostAsync)
from .extra_types import (HafasError, Log_Print, HafasError__ctor_Z384F8060)
from .types_raw_hafas_client import (Cfg, RawRequest, LocMatchRequest, RawResult, RawMatch, RawCommon, RawLoc, TripSearchRequest, RawOutCon, JourneyDetailsRequest, RawJny, StationBoardRequest, ReconstructionRequest, JourneyMatchRequest, LocGeoPosRequest, LocGeoReachRequest, RawPos, LocDetailsRequest, JourneyGeoPosRequest, HimSearchRequest, RawHim, LineMatchRequest, RawLine, ServerInfoRequest, SearchOnTripRequest, SvcReq, RawResponse, SvcRes)

_A_ = TypeVar("_A_")

def expr_312() -> TypeInfo:
    return class_type("FsHafas.Api.HafasRawClient", None, HafasRawClient)


class HafasRawClient:
    def __init__(self, endpoint: str, salt: str, cfg: Cfg, base_request: RawRequest) -> None:
        self.endpoint = endpoint
        self.salt = salt
        self.cfg = cfg
        self.base_request = base_request
        self.http_client = HttpClient__ctor()
    

HafasRawClient_reflection = expr_312

def HafasRawClient__ctor_6D5FF7F7(endpoint: str, salt: str, cfg: Cfg, base_request: RawRequest) -> HafasRawClient:
    return HafasRawClient(endpoint, salt, cfg, base_request)


def HafasRawClient__Dispose(__: HafasRawClient) -> None:
    HttpClient__Dispose(__.http_client)


def HafasRawClient__AsyncLocMatch_781FF3B0(__: HafasRawClient, lang: str, loc_match_request: LocMatchRequest) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawLoc]]]]:
    def arrow_314(__: HafasRawClient=__, lang: str=lang, loc_match_request: LocMatchRequest=loc_match_request) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawLoc]]]]:
        def arrow_313(_arg3: RawResult) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawLoc]]]]:
            res : RawResult = _arg3
            match_value : Optional[RawMatch] = res.match
            if match_value is not None:
                match : RawMatch = match_value
                return singleton.Return((res.common, res, match.loc_l))
            
            else: 
                return singleton.Return((None, None, None))
            
        
        return singleton.Bind(HafasRawClient__asyncPost_737B0FC(__, HafasRawClient__makeRequest(__, "LocMatch", lang, loc_match_request)), arrow_313)
    
    return singleton.Delay(arrow_314)


def HafasRawClient__AsyncTripSearch_Z4D007EE(__: HafasRawClient, lang: str, trip_search_request: TripSearchRequest) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawOutCon]]]]:
    def arrow_316(__: HafasRawClient=__, lang: str=lang, trip_search_request: TripSearchRequest=trip_search_request) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawOutCon]]]]:
        def arrow_315(_arg4: RawResult) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawOutCon]]]]:
            res : RawResult = _arg4
            return singleton.Return((res.common, res, res.out_con_l))
        
        return singleton.Bind(HafasRawClient__asyncPost_737B0FC(__, HafasRawClient__makeRequest(__, "TripSearch", lang, trip_search_request)), arrow_315)
    
    return singleton.Delay(arrow_316)


def HafasRawClient__AsyncJourneyDetails_73FB16B1(__: HafasRawClient, lang: str, journey_details_request: JourneyDetailsRequest) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[RawJny]]]:
    def arrow_318(__: HafasRawClient=__, lang: str=lang, journey_details_request: JourneyDetailsRequest=journey_details_request) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[RawJny]]]:
        def arrow_317(_arg5: RawResult) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[RawJny]]]:
            res : RawResult = _arg5
            return singleton.Return((res.common, res, res.journey))
        
        return singleton.Bind(HafasRawClient__asyncPost_737B0FC(__, HafasRawClient__makeRequest(__, "JourneyDetails", lang, journey_details_request)), arrow_317)
    
    return singleton.Delay(arrow_318)


def HafasRawClient__AsyncStationBoard_3FCEEA3(__: HafasRawClient, lang: str, station_board_request: StationBoardRequest) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawJny]]]]:
    def arrow_320(__: HafasRawClient=__, lang: str=lang, station_board_request: StationBoardRequest=station_board_request) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawJny]]]]:
        def arrow_319(_arg6: RawResult) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawJny]]]]:
            res : RawResult = _arg6
            return singleton.Return((res.common, res, res.jny_l))
        
        return singleton.Bind(HafasRawClient__asyncPost_737B0FC(__, HafasRawClient__makeRequest(__, "StationBoard", lang, station_board_request)), arrow_319)
    
    return singleton.Delay(arrow_320)


def HafasRawClient__AsyncReconstruction_Z19A33557(__: HafasRawClient, lang: str, reconstruction_request: ReconstructionRequest) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawOutCon]]]]:
    def arrow_322(__: HafasRawClient=__, lang: str=lang, reconstruction_request: ReconstructionRequest=reconstruction_request) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawOutCon]]]]:
        def arrow_321(_arg7: RawResult) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawOutCon]]]]:
            res : RawResult = _arg7
            return singleton.Return((res.common, res, res.out_con_l))
        
        return singleton.Bind(HafasRawClient__asyncPost_737B0FC(__, HafasRawClient__makeRequest(__, "Reconstruction", lang, reconstruction_request)), arrow_321)
    
    return singleton.Delay(arrow_322)


def HafasRawClient__AsyncJourneyMatch_Z61E31480(__: HafasRawClient, lang: str, journey_match_request: JourneyMatchRequest) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawJny]]]]:
    def arrow_324(__: HafasRawClient=__, lang: str=lang, journey_match_request: JourneyMatchRequest=journey_match_request) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawJny]]]]:
        def arrow_323(_arg8: RawResult) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawJny]]]]:
            res : RawResult = _arg8
            return singleton.Return((res.common, res, res.jny_l))
        
        return singleton.Bind(HafasRawClient__asyncPost_737B0FC(__, HafasRawClient__makeRequest(__, "JourneyMatch", lang, journey_match_request)), arrow_323)
    
    return singleton.Delay(arrow_324)


def HafasRawClient__AsyncLocGeoPos_3C0E4562(__: HafasRawClient, lang: str, loc_geo_pos_request: LocGeoPosRequest) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawLoc]]]]:
    def arrow_326(__: HafasRawClient=__, lang: str=lang, loc_geo_pos_request: LocGeoPosRequest=loc_geo_pos_request) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawLoc]]]]:
        def arrow_325(_arg9: RawResult) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawLoc]]]]:
            res : RawResult = _arg9
            return singleton.Return((res.common, res, res.loc_l))
        
        return singleton.Bind(HafasRawClient__asyncPost_737B0FC(__, HafasRawClient__makeRequest(__, "LocGeoPos", lang, loc_geo_pos_request)), arrow_325)
    
    return singleton.Delay(arrow_326)


def HafasRawClient__AsyncLocGeoReach_320A81B3(__: HafasRawClient, lang: str, loc_geo_reach_request: LocGeoReachRequest) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], List[RawPos]]]:
    def arrow_328(__: HafasRawClient=__, lang: str=lang, loc_geo_reach_request: LocGeoReachRequest=loc_geo_reach_request) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], List[RawPos]]]:
        def arrow_327(_arg10: RawResult) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], List[RawPos]]]:
            res : RawResult = _arg10
            match_value : Optional[List[RawPos]] = res.pos_l
            if match_value is not None:
                pos_l : List[RawPos] = match_value
                return singleton.Return((res.common, res, pos_l))
            
            else: 
                return singleton.Return((None, None, []))
            
        
        return singleton.Bind(HafasRawClient__asyncPost_737B0FC(__, HafasRawClient__makeRequest(__, "LocGeoReach", lang, loc_geo_reach_request)), arrow_327)
    
    return singleton.Delay(arrow_328)


def HafasRawClient__AsyncLocDetails_2FCF6141(__: HafasRawClient, lang: str, loc_details_request: LocDetailsRequest) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[RawLoc]]]:
    def arrow_330(__: HafasRawClient=__, lang: str=lang, loc_details_request: LocDetailsRequest=loc_details_request) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[RawLoc]]]:
        def arrow_329(_arg11: RawResult) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[RawLoc]]]:
            res : RawResult = _arg11
            match_value : Optional[List[RawLoc]] = res.loc_l
            (pattern_matching_result, loc_l_1) = (None, None)
            if match_value is not None:
                if len(match_value) > 0:
                    pattern_matching_result = 0
                    loc_l_1 = match_value
                
                else: 
                    pattern_matching_result = 1
                
            
            else: 
                pattern_matching_result = 1
            
            if pattern_matching_result == 0:
                return singleton.Return((res.common, res, loc_l_1[0]))
            
            elif pattern_matching_result == 1:
                return singleton.Return((None, None, None))
            
        
        return singleton.Bind(HafasRawClient__asyncPost_737B0FC(__, HafasRawClient__makeRequest(__, "LocDetails", lang, loc_details_request)), arrow_329)
    
    return singleton.Delay(arrow_330)


def HafasRawClient__AsyncJourneyGeoPos_Z6DB7B6AE(__: HafasRawClient, lang: str, journey_geo_pos_request: JourneyGeoPosRequest) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawJny]]]]:
    def arrow_332(__: HafasRawClient=__, lang: str=lang, journey_geo_pos_request: JourneyGeoPosRequest=journey_geo_pos_request) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawJny]]]]:
        def arrow_331(_arg12: RawResult) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawJny]]]]:
            res : RawResult = _arg12
            return singleton.Return((res.common, res, res.jny_l))
        
        return singleton.Bind(HafasRawClient__asyncPost_737B0FC(__, HafasRawClient__makeRequest(__, "JourneyGeoPos", lang, journey_geo_pos_request)), arrow_331)
    
    return singleton.Delay(arrow_332)


def HafasRawClient__AsyncHimSearch_Z6033479F(__: HafasRawClient, lang: str, him_search_request: HimSearchRequest) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawHim]]]]:
    def arrow_334(__: HafasRawClient=__, lang: str=lang, him_search_request: HimSearchRequest=him_search_request) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawHim]]]]:
        def arrow_333(_arg13: RawResult) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawHim]]]]:
            res : RawResult = _arg13
            return singleton.Return((res.common, res, res.msg_l))
        
        return singleton.Bind(HafasRawClient__asyncPost_737B0FC(__, HafasRawClient__makeRequest(__, "HimSearch", lang, him_search_request)), arrow_333)
    
    return singleton.Delay(arrow_334)


def HafasRawClient__AsyncLineMatch_Z69AA7EA2(__: HafasRawClient, lang: str, line_match_request: LineMatchRequest) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawLine]]]]:
    def arrow_336(__: HafasRawClient=__, lang: str=lang, line_match_request: LineMatchRequest=line_match_request) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawLine]]]]:
        def arrow_335(_arg14: RawResult) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawLine]]]]:
            res : RawResult = _arg14
            return singleton.Return((res.common, res, res.line_l))
        
        return singleton.Bind(HafasRawClient__asyncPost_737B0FC(__, HafasRawClient__makeRequest(__, "LineMatch", lang, line_match_request)), arrow_335)
    
    return singleton.Delay(arrow_336)


def HafasRawClient__AsyncServerInfo_Z684EE398(__: HafasRawClient, lang: str, server_info_request: ServerInfoRequest) -> Async[Tuple[Optional[RawCommon], Optional[RawResult]]]:
    def arrow_338(__: HafasRawClient=__, lang: str=lang, server_info_request: ServerInfoRequest=server_info_request) -> Async[Tuple[Optional[RawCommon], Optional[RawResult]]]:
        def arrow_337(_arg15: RawResult) -> Async[Tuple[Optional[RawCommon], Optional[RawResult]]]:
            res : RawResult = _arg15
            return singleton.Return((res.common, res))
        
        return singleton.Bind(HafasRawClient__asyncPost_737B0FC(__, HafasRawClient__makeRequest(__, "ServerInfo", lang, server_info_request)), arrow_337)
    
    return singleton.Delay(arrow_338)


def HafasRawClient__AsyncSearchOnTrip_5D40A373(__: HafasRawClient, lang: str, search_on_trip_request: SearchOnTripRequest) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawOutCon]]]]:
    def arrow_340(__: HafasRawClient=__, lang: str=lang, search_on_trip_request: SearchOnTripRequest=search_on_trip_request) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawOutCon]]]]:
        def arrow_339(_arg16: RawResult) -> Async[Tuple[Optional[RawCommon], Optional[RawResult], Optional[List[RawOutCon]]]]:
            res : RawResult = _arg16
            return singleton.Return((res.common, res, res.out_con_l))
        
        return singleton.Bind(HafasRawClient__asyncPost_737B0FC(__, HafasRawClient__makeRequest(__, "SearchOnTrip", lang, search_on_trip_request)), arrow_339)
    
    return singleton.Delay(arrow_340)


def HafasRawClient__toException_ZAAE3ECF(this: HafasRawClient, ex: HafasError) -> HafasError:
    return ex


def HafasRawClient__log(this: HafasRawClient, msg: str, o: _A_) -> None:
    Log_Print(msg, o)


def HafasRawClient__makeRequest(this: HafasRawClient, meth: str, lang: str, parameters: Any=None) -> RawRequest:
    input_record : RawRequest = this.base_request
    return RawRequest(lang, [SvcReq(this.cfg, meth, parameters)], input_record.client, input_record.ext, input_record.ver, input_record.auth)


def HafasRawClient__asyncPost_737B0FC(this: HafasRawClient, request: RawRequest) -> Async[RawResult]:
    json : str = encode(request)
    HafasRawClient__log(this, "request:", json)
    def arrow_356(this: HafasRawClient=this, request: RawRequest=request) -> Async[RawResult]:
        def arrow_355(_arg1: str) -> Async[RawResult]:
            result : str = _arg1
            HafasRawClient__log(this, "response:", result)
            def arrow_351(__unit: Any=None) -> Async[RawResult]:
                if len(result) == 0:
                    def arrow_341(__unit: Any=None) -> RawResult:
                        raise Exception("invalid response")
                    
                    return singleton.Return(arrow_341())
                
                else: 
                    response : RawResponse = decode(result)
                    svc_res_l : List[SvcRes] = default_arg(response.svc_res_l, [])
                    if len(svc_res_l) == 1:
                        svc_res : SvcRes = svc_res_l[0]
                        match_value : Tuple[Optional[str], Optional[str], Optional[RawResult]] = (svc_res.err, svc_res.err_txt, svc_res.res)
                        (pattern_matching_result, err_1, err_txt_1) = (None, None, None)
                        if match_value[0] is not None:
                            if match_value[1] is not None:
                                def arrow_345(__unit: Any=None) -> bool:
                                    err_txt : str = match_value[1]
                                    return match_value[0] != "OK"
                                
                                if arrow_345():
                                    pattern_matching_result = 0
                                    err_1 = match_value[0]
                                    err_txt_1 = match_value[1]
                                
                                else: 
                                    pattern_matching_result = 1
                                
                            
                            else: 
                                pattern_matching_result = 1
                            
                        
                        else: 
                            pattern_matching_result = 1
                        
                        if pattern_matching_result == 0:
                            def arrow_342(__unit: Any=None) -> RawResult:
                                raise HafasError__ctor_Z384F8060(err_1, err_txt_1)
                            
                            return singleton.Return(arrow_342())
                        
                        elif pattern_matching_result == 1:
                            (pattern_matching_result_1, err_3) = (None, None)
                            if match_value[0] is not None:
                                if match_value[0] != "OK":
                                    pattern_matching_result_1 = 0
                                    err_3 = match_value[0]
                                
                                else: 
                                    pattern_matching_result_1 = 1
                                
                            
                            else: 
                                pattern_matching_result_1 = 1
                            
                            if pattern_matching_result_1 == 0:
                                def arrow_343(__unit: Any=None) -> RawResult:
                                    raise HafasError__ctor_Z384F8060(err_3, err_3)
                                
                                return singleton.Return(arrow_343())
                            
                            elif pattern_matching_result_1 == 1:
                                if match_value[2] is not None:
                                    res : RawResult = match_value[2]
                                    return singleton.Return(res)
                                
                                else: 
                                    def arrow_344(__unit: Any=None) -> RawResult:
                                        raise Exception("invalid response")
                                    
                                    return singleton.Return(arrow_344())
                                
                            
                        
                    
                    else: 
                        match_value_1 : Tuple[Optional[str], Optional[str]] = (response.err, response.err_txt)
                        def arrow_347(__unit: Any=None) -> Async[RawResult]:
                            err_4 : str = match_value_1[0]
                            err_txt_2 : str = match_value_1[1]
                            def arrow_346(__unit: Any=None) -> RawResult:
                                raise HafasError__ctor_Z384F8060(err_4, err_txt_2)
                            
                            return singleton.Return(arrow_346())
                        
                        def arrow_349(__unit: Any=None) -> Async[RawResult]:
                            err_5 : str = match_value_1[0]
                            def arrow_348(__unit: Any=None) -> RawResult:
                                raise HafasError__ctor_Z384F8060(err_5, err_5)
                            
                            return singleton.Return(arrow_348())
                        
                        def arrow_350(__unit: Any=None) -> RawResult:
                            raise Exception("invalid response")
                        
                        return (arrow_347() if (match_value_1[1] is not None) else arrow_349()) if (match_value_1[0] is not None) else singleton.Return(arrow_350())
                    
                
            
            def arrow_354(_arg2: Exception) -> Async[RawResult]:
                if isinstance(_arg2, HafasError):
                    def arrow_352(__unit: Any=None) -> RawResult:
                        raise HafasRawClient__toException_ZAAE3ECF(this, _arg2)
                    
                    return singleton.Return(arrow_352())
                
                else: 
                    arg10_1 : str = str(_arg2)
                    to_console(printf("error: %s"))(arg10_1)
                    def arrow_353(__unit: Any=None) -> RawResult:
                        raise Exception("invalid response")
                    
                    return singleton.Return(arrow_353())
                
            
            return singleton.TryWith(singleton.Delay(arrow_351), arrow_354)
        
        return singleton.Bind(HttpClient__PostAsync(this.http_client, this.endpoint, this.salt, json), arrow_355)
    
    return singleton.Delay(arrow_356)





from __future__ import annotations
from typing import (Any, Optional)
from ...fable_library.list import (FSharpList, last)
from ...fable_library.map import (of_seq, FSharpMap__TryFind)
from ...fable_library.reg_exp import replace
from ...fable_library.string import (substring, to_text, printf, to_console, replace as replace_1)
from ...fable_library.util import compare_primitives
from ...fable_simple_json_python.json_converter import Convert_serialize
from ...fable_simple_json_python.simple_json import (SimpleJson_toString, SimpleJson_mapKeysByPath, SimpleJson_parseNative)
from ...fable_simple_json_python.type_info_converter import create_type_info
from ...fs_hafas_python.types_raw_hafas_client import (SvcReq, Cfg_reflection, LocMatchRequest, LocMatchRequest_reflection, TripSearchRequest, TripSearchRequest_reflection, JourneyDetailsRequest, JourneyDetailsRequest_reflection, StationBoardRequest, StationBoardRequest_reflection, ReconstructionRequest, ReconstructionRequest_reflection, JourneyMatchRequest, JourneyMatchRequest_reflection, LocGeoPosRequest, LocGeoPosRequest_reflection, LocGeoReachRequest, LocGeoReachRequest_reflection, LocDetailsRequest, LocDetailsRequest_reflection, JourneyGeoPosRequest, JourneyGeoPosRequest_reflection, HimSearchRequest, HimSearchRequest_reflection, LineMatchRequest, LineMatchRequest_reflection, SearchOnTripRequest, SearchOnTripRequest_reflection, RawRequest, RawRequestClient_reflection, RawRequestAuth_reflection)

class ObjectExpr300:
    @property
    def Compare(self) -> Any:
        def arrow_299(x: str, y: str) -> int:
            return compare_primitives(x, y)
        
        return arrow_299
    

replacements : Any = of_seq([("a_tzoffset", "aTZOffset"), ("crd_enc_yx", "crdEncYX"), ("d_tzoffset", "dTZOffset"), ("d_trn_cmp_sx", "dTrnCmpSX"), ("get_ist", "getIST"), ("get_iv", "getIV"), ("get_pois", "getPOIs"), ("get_pt", "getPT"), ("only_rt", "onlyRT"), ("planrt_ts", "planrtTS"), ("poly_xl", "polyXL"), ("show_arslink", "showARSLink")], ObjectExpr300())

def undashify(input: str) -> str:
    def arrow_301(m: Any, input: str=input) -> str:
        return substring(m[0], 1, 1).upper()
    
    return replace(input, "_[a-z]", arrow_301)


def to_undashed(xs: FSharpList[str]) -> Optional[str]:
    v : str = last(xs)
    match_value : Optional[str] = FSharpMap__TryFind(replacements, v)
    if match_value is None:
        return undashify(v)
    
    else: 
        return match_value
    


def encode_svc_req(svc_req: SvcReq) -> str:
    cfg : str = Convert_serialize(svc_req.cfg, create_type_info(Cfg_reflection()))
    req : str
    match_value : Any = svc_req.req
    req = Convert_serialize(match_value, create_type_info(LocMatchRequest_reflection())) if isinstance(match_value, LocMatchRequest) else (Convert_serialize(match_value, create_type_info(TripSearchRequest_reflection())) if isinstance(match_value, TripSearchRequest) else (Convert_serialize(match_value, create_type_info(JourneyDetailsRequest_reflection())) if isinstance(match_value, JourneyDetailsRequest) else (Convert_serialize(match_value, create_type_info(StationBoardRequest_reflection())) if isinstance(match_value, StationBoardRequest) else (Convert_serialize(match_value, create_type_info(ReconstructionRequest_reflection())) if isinstance(match_value, ReconstructionRequest) else (Convert_serialize(match_value, create_type_info(JourneyMatchRequest_reflection())) if isinstance(match_value, JourneyMatchRequest) else (Convert_serialize(match_value, create_type_info(LocGeoPosRequest_reflection())) if isinstance(match_value, LocGeoPosRequest) else (Convert_serialize(match_value, create_type_info(LocGeoReachRequest_reflection())) if isinstance(match_value, LocGeoReachRequest) else (Convert_serialize(match_value, create_type_info(LocDetailsRequest_reflection())) if isinstance(match_value, LocDetailsRequest) else (Convert_serialize(match_value, create_type_info(JourneyGeoPosRequest_reflection())) if isinstance(match_value, JourneyGeoPosRequest) else (Convert_serialize(match_value, create_type_info(HimSearchRequest_reflection())) if isinstance(match_value, HimSearchRequest) else (Convert_serialize(match_value, create_type_info(LineMatchRequest_reflection())) if isinstance(match_value, LineMatchRequest) else (Convert_serialize(match_value, create_type_info(SearchOnTripRequest_reflection())) if isinstance(match_value, SearchOnTripRequest) else "{}"))))))))))))
    return to_text(printf("{\"cfg\":%s, \"meth\":\"%s\", \"req\":%s}"))(cfg)(svc_req.meth)(req)


def encode(request: RawRequest) -> str:
    try: 
        svcreql : str = ("[" + encode_svc_req(request.svc_req_l[0])) + "]"
        client : str = Convert_serialize(request.client, create_type_info(RawRequestClient_reflection()))
        auth : str = Convert_serialize(request.auth, create_type_info(RawRequestAuth_reflection()))
        def arrow_302(xs: FSharpList[str]) -> Optional[str]:
            return to_undashed(xs)
        
        return SimpleJson_toString(SimpleJson_mapKeysByPath(arrow_302, SimpleJson_parseNative(replace_1(replace_1(to_text(printf("{\"lang\":\"%s\", \"svcReqL\":%s, \"client\":%s, \"ext\":\"%s\", \"ver\":\"%s\", \"auth\":%s}"))(request.lang)(svcreql)(client)(request.ext)(request.ver)(auth), ", \"meta\": null", ""), ",\"name\":null", ""))))
    
    except Exception as e:
        arg10_1 : str = str(e)
        to_console(printf("error encode: %s"))(arg10_1)
        raise Exception(str(e))
    





from abc import abstractmethod
from typing import Any, Callable, Generic, List, Optional, Protocol, Tuple, TypeVar

from .choice import (
    Choice_tryValueIfChoice1Of2,  # type: ignore
    Choice_tryValueIfChoice2Of2,
    FSharpChoice_2,
)
from .option import value
from .util import IDisposable

_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_T_contra = TypeVar("_T_contra", contravariant=True)
_U = TypeVar("_U")


class IObserver(Protocol, Generic[_T_contra]):
    @abstractmethod
    def OnNext(self, __value: _T_contra) -> None:
        ...

    @abstractmethod
    def OnError(self, __error: Exception) -> None:
        ...

    @abstractmethod
    def OnCompleted(self) -> None:
        ...


def _noop(__arg: Any = None) -> None:
    pass


class Observer(IObserver[_T]):
    def __init__(
        self,
        on_next: Callable[[_T], None],
        on_error: Optional[Callable[[Exception], None]] = None,
        on_completed: Optional[Callable[[], None]] = None,
    ) -> None:
        self._on_next = on_next
        self._on_error = on_error or _noop
        self._on_completed = on_completed or _noop

    def OnError(self, error: Exception) -> None:
        return self._on_error(error)

    def OnNext(self, value: _T) -> None:
        return self._on_next(value)

    def OnCompleted(self) -> None:
        return self._on_completed()


class IObservable(Protocol, Generic[_T_co]):
    @abstractmethod
    def Subscribe(self, __obs: IObserver[_T_co]) -> IDisposable:
        ...


class Observable(IObservable[_T]):
    def __init__(self, subscribe: Callable[[IObserver[_T]], IDisposable]) -> None:
        self.subscribe = subscribe

    def Subscribe(self, obv: IObserver[_T]) -> IDisposable:
        return self.subscribe(obv)


def subscribe(callback: Callable[[_T], None], source: IObservable[_T]) -> IDisposable:
    return source.Subscribe(Observer(callback))


def add(callback: Callable[[_T], None], source: IObservable[_T]) -> IDisposable:
    return source.Subscribe(Observer(callback))


def protect(
    f: Callable[[], _T],
    succeed: Callable[[_T], None],
    fail: Callable[[Exception], None],
):
    try:
        return succeed(f())
    except Exception as e:
        fail(e)


def choose(
    chooser: Callable[[_T], Optional[_U]], source: IObservable[_T]
) -> IObservable[_U]:
    def subscribe(observer: IObserver[_U]):
        def on_next(t: _T) -> None:
            def success(u: Optional[_U]) -> None:
                if u is not None:
                    observer.OnNext(value(u))

            return protect(lambda: chooser(t), success, observer.OnError)

        obv = Observer(on_next, observer.OnError, observer.OnCompleted)
        return source.Subscribe(obv)

    return Observable(subscribe)


def filter(predicate: Callable[[_T], bool], source: IObservable[_T]) -> IObservable[_T]:
    return choose(lambda x: x if predicate(x) else None, source)


def map(mapping: Callable[[_T], _U], source: IObservable[_T]) -> IObservable[_U]:
    def subscribe(observer: IObserver[_U]) -> IDisposable:
        def on_next(value: _T) -> None:
            return protect(lambda: mapping(value), observer.OnNext, observer.OnError)

        obv = Observer(on_next, observer.OnError, observer.OnCompleted)
        return source.Subscribe(obv)

    return Observable(subscribe)


def merge(source1: IObservable[_T], source2: IObservable[_T]) -> IObservable[_T]:
    def subscribe(observer: IObserver[_T]) -> IDisposable:
        stopped = False
        completed1 = False
        completed2 = False

        def on_next(value: _T) -> None:
            if stopped:
                return
            observer.OnNext(value)

        def on_error(error: Exception) -> None:
            nonlocal stopped
            if stopped:
                return
            stopped = True
            observer.OnError(error)

        def on_completed1() -> None:
            nonlocal completed1, stopped
            if stopped:
                return

            completed1 = True
            if completed2:
                stopped = True
                observer.OnCompleted()

        obv1 = Observer(on_next, on_error, on_completed1)
        h1 = source1.Subscribe(obv1)

        def on_completed2() -> None:
            nonlocal completed2, stopped
            if stopped:
                return

            completed2 = True
            if completed1:
                stopped = True
                observer.OnCompleted()

        obv2 = Observer(on_next, on_error, on_completed2)
        h2 = source2.Subscribe(obv2)

        def dispose() -> None:
            h1.Dispose()
            h2.Dispose()

        return IDisposable.create(dispose)

    return Observable(subscribe)


def pairwise(source: IObservable[_T]) -> IObservable[Tuple[_T, _T]]:
    def subscribe(observer: IObserver[Tuple[_T, _T]]) -> IDisposable:
        last: Optional[_T] = None

        def on_next(value: _T) -> None:
            nonlocal last
            if last is not None:
                observer.OnNext((last, value))

            last = value

        obv = Observer(on_next, observer.OnError, observer.OnCompleted)
        return source.Subscribe(obv)

    return Observable(subscribe)


def partition(
    predicate: Callable[[_T], bool], source: IObservable[_T]
) -> Tuple[IObservable[_T], IObservable[_T]]:
    return (filter(predicate, source), filter(lambda x: not predicate(x), source))


def scan(
    collector: Callable[[_U, _T], _U], state: _U, source: IObservable[_T]
) -> IObservable[_U]:
    def subscribe(observer: IObserver[_U]) -> IDisposable:
        def on_next(t: _T) -> None:
            def success(u: _U) -> None:
                nonlocal state
                state = u
                observer.OnNext(u)

            protect(lambda: collector(state, t), success, observer.OnError)

        obv = Observer(on_next, observer.OnError, observer.OnCompleted)
        return source.Subscribe(obv)

    return Observable(subscribe)


def split(
    splitter: Callable[[_T], FSharpChoice_2], source: IObservable[_T]
) -> Tuple[IObservable[_T], IObservable[_T]]:
    return (
        choose(lambda v: Choice_tryValueIfChoice1Of2(splitter(v)), source),
        choose(lambda v: Choice_tryValueIfChoice2Of2(splitter(v)), source),
    )
